try:
    import matlab.engine
    matlab_available = True
except:
    matlab_available = False
# To install matlab engine, go to folder /usr/local/MATLAB/R2017a/extern/engines/python
# and run terminal command: sudo python setup.py install
import argparse
from time import sleep, time
import os
import tempfile
import shutil
import copy
from multiprocessing import Process

import numpy as np

from openEPhys_DACQ.package_configuration import package_config
from openEPhys_DACQ import NWBio
from openEPhys_DACQ.createAxonaData import createAxonaData_for_NWBfile
from openEPhys_DACQ import HelperFunctions as hfunct
from openEPhys_DACQ.KlustaKwikWrapper import applyKlustaKwik_on_spike_data_tet
from openEPhys_DACQ.TrackingDataProcessing import (remove_tracking_data_jumps,
                                                   iteratively_combine_multicamera_data_for_recording)

def use_binary_pos(filename, postprocess=False, maxjump=25):
    """
    Copies binary position data into tracking data
    Apply postprocessing with postprocess=True
    """
    recordingKey = NWBio.get_recordingKey(filename)
    # Load timestamps and position data
    with h5py.File(filename, 'r+') as h5file:
        timestamps = np.array(h5file['acquisition']['timeseries'][recordingKey]['events']['binary1']['timestamps'])
        xy = np.array(h5file['acquisition']['timeseries'][recordingKey]['events']['binary1']['data'][:,:2])
    data = {'xy': xy, 'timestamps': timestamps}
    # Construct data into repository familiar posdata format (single array with times in first column)
    posdata = np.append(data['timestamps'][:,None], data['xy'].astype(np.float64), axis=1)
    # Add NaNs for second LED if missing
    if posdata.shape[1] < 5:
        nanarray = np.zeros(data['xy'].shape, dtype=np.float64)
        nanarray[:] = np.nan
        posdata = np.append(posdata, nanarray, axis=1)
    # Postprocess the data if requested
    if postprocess:
        posdata = remove_tracking_data_jumps(posdata, maxjump)
    # Save data to ProcessedPos position in NWB file
    NWBio.save_tracking_data(filename, posdata, ProcessedPos=True, overwrite=True)


def process_tracking_data(filename, save_to_file=False, verbose=False):
    if verbose:
        print('Processing tracking data for {}'.format(filename))

    # Get CameraSettings
    try:
        CameraSettings = NWBio.load_settings(filename, '/CameraSettings/')
    except KeyError as e:
        raise Exception('Could not load CameraSettings for {}'.format(filename)
                        + '\nLikely older version on file format.')
    cameraIDs = sorted(CameraSettings['CameraSpecific'].keys())
    # Get global clock timestamps
    OE_GC_times = NWBio.load_GlobalClock_timestamps(filename)
    # Get arena_size
    arena_size = NWBio.load_settings(filename, '/General/arena_size/')
    # Load position data for all cameras
    posdatas = []
    for cameraID in cameraIDs:
        posdatas.append(NWBio.load_raw_tracking_data(filename, cameraID))
    ProcessedPos = iteratively_combine_multicamera_data_for_recording(
        CameraSettings, arena_size, posdatas, OE_GC_times, verbose=verbose)
    if save_to_file:
        # Save corrected data to file
        NWBio.save_tracking_data(filename, ProcessedPos, ProcessedPos=True, overwrite=True)

    return ProcessedPos


def recompute_tracking_data_for_all_files_in_directory_tree(root_path, verbose=False):

    # Commence directory walk
    processeses = []
    for dir_name, subdirList, fileList in os.walk(root_path):
        for fname in fileList:
            fpath = os.path.join(dir_name, fname)
            if fname == 'experiment_1.nwb':
                p = Process(target=process_tracking_data, args=(fpath,),
                            kwargs={'save_to_file': True, 'verbose': verbose})
                p.start()
                processeses.append(p)
    for p in processeses:
        p.join()


def lowpass_and_downsample_channel(
        fpath, chan, original_sampling_rate, target_sampling_rate):
    data = NWBio.load_continuous_as_array(fpath, chan)['continuous'].squeeze()
    data = hfunct.lowpass_and_downsample(data, original_sampling_rate, target_sampling_rate)

    return data


def lowpass_and_downsample_channels(
        fpath, channels, original_sampling_rate, target_sampling_rate):
    # Load, lowpass filter and downsample each channel in a separate process
    multiprocessor = hfunct.multiprocess()
    for chan in channels:
        hfunct.proceed_when_enough_memory_available(percent=0.50)
        print(hfunct.time_string() + ' Starting lowpass_and_downsample_channel for chan ' + str(chan))
        multiprocessor.run(lowpass_and_downsample_channel,
                           args=(fpath, chan, original_sampling_rate, target_sampling_rate))
        sleep(4)
    # Collect processed data from multiprocess class
    out = multiprocessor.results()

    return out


def lowpass_and_downsample_channel_on_each_tetrode(
        fpath, original_sampling_rate, target_sampling_rate, n_tetrodes, badChans):
    processed_chans = []
    processed_tets = []
    for n_tet in range(n_tetrodes):
        chans = hfunct.tetrode_channels(n_tet)
        chan = []
        for c in chans:
            if c not in badChans:
                chan.append(c)
        if len(chan) > 0:
            processed_chans.append(chan[0])
            processed_tets.append(n_tet)
    processed_data_list = lowpass_and_downsample_channels(
        fpath, processed_chans, original_sampling_rate, target_sampling_rate)
    new_data_size = processed_data_list[0].size
    processed_data_array = np.zeros((new_data_size, n_tetrodes), dtype=np.int16)
    for n_tet, processed_data in zip(processed_tets, processed_data_list):
        processed_data_array[:, n_tet] = np.int16(processed_data)

    return processed_data_array, processed_chans


def lowpass_and_downsample_AUX_data(fpath, n_tetrodes, original_sampling_rate, target_sampling_rate):
    aux_chan_list = NWBio.list_AUX_channels(fpath, n_tetrodes)
    processed_data_list = lowpass_and_downsample_channels(
        fpath, aux_chan_list, original_sampling_rate, target_sampling_rate)
    downsampled_AUX = np.concatenate([x[:, None] for x in processed_data_list], axis=1)
    downsampled_AUX = downsampled_AUX.astype(np.int16)

    return downsampled_AUX


def downsample_raw_timestamps(fpath, downsample_factor):
    return NWBio.load_raw_data_timestamps_as_array(fpath)[::downsample_factor]


def create_downsampled_data(fpath, n_tetrodes=32, downsample_factor=10):
    # Get original sampling rate and compute target rate based on downsampling factor
    original_sampling_rate = NWBio.OpenEphys_SamplingRate()
    target_sampling_rate = int(NWBio.OpenEphys_SamplingRate() / downsample_factor)
    # Get list of bad channels
    badChans = NWBio.listBadChannels(fpath)
    # Get downsampled data
    downsampled_data, used_chans = lowpass_and_downsample_channel_on_each_tetrode(
        fpath, original_sampling_rate, target_sampling_rate, n_tetrodes, badChans)
    downsampled_AUX = lowpass_and_downsample_AUX_data(fpath, n_tetrodes, original_sampling_rate, target_sampling_rate)
    downsampled_timestamps = downsample_raw_timestamps(fpath, downsample_factor)
    # Ensure timestamps and downsampled data have same number of samples
    assert downsampled_timestamps.size == downsampled_data.shape[0], \
        'Downsampled timestamps does not match number of samples in downsampled continuous data.'
    # Create downsampling_info
    downsampling_info = {'original_sampling_rate': original_sampling_rate,
                         'downsampled_sampling_rate': target_sampling_rate,
                         'downsampled_channels': np.array(used_chans)}
    # Save downsampled data to disk
    NWBio.save_downsampled_data_to_disk(
        fpath, downsampled_data, downsampled_timestamps, downsampled_AUX, downsampling_info)


def delete_raw_data(fpath, only_if_downsampled_data_available=True):
    print(hfunct.time_string() + ' Deleting raw data in ' + fpath)
    NWBio.delete_raw_data(fpath, only_if_downsampled_data_available=only_if_downsampled_data_available)
    print(hfunct.time_string() + ' Repacking NWB file ' + fpath)
    NWBio.repack_NWB_file(fpath, replace_original=True)


class ContinuousDataPreloader(object):
    """
    This class loads into memory and preprocesses continuous data.
    Data for specific channels can then be queried.
        channels - continuous list of channels to prepare, e.g. list(range(0,16,1)) for first 4 tetrodes
    """
    def __init__(self, OpenEphysDataPath, channels):
        self.chan_nrs = list(channels)
        # Load data
        print('Loading continuous data from NWB file.')
        data = NWBio.load_continuous_as_array(OpenEphysDataPath, self.chan_nrs)
        self.timestamps = data['timestamps']
        self.continuous = data['continuous']
        self.continuous = np.transpose(self.continuous)
        # Set bad channels to 0
        self.badChan = NWBio.listBadChannels(OpenEphysDataPath)
        self.continuous = set_bad_chan_to_0(self.continuous, self.chan_nrs, self.badChan)
        print('Loading continuous data from NWB file successful.')

    def get_chan_nrs_idx_in_continuous(self, chan_nrs_req):
        """
        Finds rows in self.continuous corresponding to chan_nrs requested
        """
        chan_nrs_idx = []
        for chan_nr in chan_nrs_req:
            chan_nrs_idx.append(self.chan_nrs.index(chan_nr))

        return chan_nrs_idx

    def prepare_referencing(self, referencing_method='other_channels'):
        """
        Prepares variables for channel specific use of method ContinuousDataPreloader.referenced
            referencing_method options: 'other_channels' or 'all_channels'
        """
        self.referencing_method = referencing_method
        if referencing_method == 'other_channels':
            self.continuous_sum = np.sum(self.continuous, axis=0)
        elif referencing_method == 'all_channels':
            self.continuous_mean = np.mean(self.continuous, axis=0).astype(np.int16)

    def referenced_continuous(self, chan_nrs_req):
        """
        Returns referenced continuous data for channels in list chan_nrs_req.
        If prepare_referencing method has not been called, calls it with default options.
        """
        if not hasattr(self, 'referencing_method'):
            self.prepare_referencing()
        chan_nrs_idx = self.get_chan_nrs_idx_in_continuous(chan_nrs_req)
        # Compute referencing array according to method
        if self.referencing_method == 'other_channels':
            continuous_other_sum = self.continuous_sum - np.sum(self.continuous[chan_nrs_idx], axis=0)
            continuous_other_mean = continuous_other_sum.astype(np.float64) / np.float64(self.continuous.shape[0])
            reference_array = np.int16(continuous_other_mean)
        elif self.referencing_method == 'all_channels':
            reference_array = self.continuous_mean
        # Extract data and apply referencing array
        data = np.zeros((len(chan_nrs_req), self.continuous.shape[1]), dtype=self.continuous.dtype)
        for idx, chan_nr in zip(chan_nrs_idx, chan_nrs_req):
            if not (chan_nr in self.badChan):
                data[chan_nrs_req.index(chan_nr), :] = self.continuous[idx, :] - reference_array

        return data

    def filter_signal(self, signal_in, filter_freqs):
        signal_out = hfunct.butter_bandpass_filter(signal_in, sampling_rate=30000.0, 
                                                   highpass_frequency=filter_freqs[0], 
                                                   lowpass_frequency=filter_freqs[1], 
                                                   filt_order=4)
        return signal_out

    def get_channels(self, chan_nrs_req, referenced=False, filter_freqs=False, no_badChan=False):
        """
        Returns an array where each row is continuous signal for en element in list chan_nrs_req.
            referenced=True - output data is referenced with method referenced_continuous
            filter_freqs=[300, 6000] - sets band-pass filtering frequency band limits
        """
        if referenced:
            data = self.referenced_continuous(chan_nrs_req)
        else:
            chan_nrs_idx = self.get_chan_nrs_idx_in_continuous(chan_nrs_req)
            data = self.continuous[chan_nrs_idx, :]
        if no_badChan:
            badChan_idx = []
            for i, chan_nr in enumerate(chan_nrs_req):
                if chan_nr in self.badChan:
                    badChan_idx.append(i)
            data = np.delete(data, badChan_idx, axis=0)
        if filter_freqs:
            for nchan in range(data.shape[0]):
                data[nchan, :] = self.filter_signal(data[nchan, :], filter_freqs)            

        return data

    def close(self):
        """
        Deletes pre-loaded data from virtual memory
        """
        if hasattr(self, 'continuous'):
            del self.continuous
        if hasattr(self, 'timestamps'):
            del self.timestamps
        if hasattr(self, 'continuous_sum'):
            del self.continuous_sum
        if hasattr(self, 'continuous_mean'):
            del self.continuous_mean

def set_bad_chan_to_0(data, chan_nrs, badChan):
    """
    Sets channels (rows in matrix data) to 0 values according to lists chan_nrs and badChan
    If badChan is an empty list, data will be unchanged
    """
    if len(badChan) > 0:
        for nchanBad in badChan:
            if nchanBad in chan_nrs:
                nchan = chan_nrs.index(nchanBad)
                data[nchan,:] = np.int16(0)

    return data

def detect_threshold_crossings_on_tetrode(continuous_tetrode_data, threshold, tooclose, detection_method='negative'):
    """
    Finds all threshold crossings on a tetrode. Returns an empty array if no threshold crossings detected
        continuous_tetrode_data - 4 x N processed continuous data array for 4 channels at N datapoints
        threshold - threshold value in microvolts
        tooclose - minimum latency allowed between spikes (in datapoints)
        detection_method - 'negative', 'positive' or 'both' - the polarity of threshold crossing detection
    """
    threshold_int16 = np.int16(np.round(threshold / 0.195))
    # Find threshold crossings for each channel
    spike_indices = np.array([], dtype=np.int64)
    for nchan in range(continuous_tetrode_data.shape[0]):
        if detection_method == 'negative':
            tmp = continuous_tetrode_data[nchan,:] < -threshold_int16
        elif detection_method == 'positive':
            tmp = continuous_tetrode_data[nchan,:] > threshold_int16
        elif detection_method == 'both':
            tmp = np.abs(continuous_tetrode_data[nchan,:]) > threshold_int16
        spike_indices = np.append(spike_indices, np.where(tmp)[0])
    # Sort the spike indices
    if len(spike_indices) > 0: 
        spike_indices = np.sort(spike_indices)
    # Remove duplicates based on temporal proximity
    if len(spike_indices) > 0:
        spike_diff = np.append(np.array([0]),np.diff(spike_indices))
        tooclose_idx = spike_diff < tooclose
        spike_indices = np.delete(spike_indices, np.where(tooclose_idx)[0])

    return spike_indices

def extract_spikes_from_tetrode(continuous_tetrode_data, spike_indices, waveform_length=[6, 34]):
    """
    Extracts spike waveforms from continuous_data based on spike timestamps
        continuous_tetrode_data - 4 x N processed continuous data array for 4 channels at N datapoints
        spike_indices - indices for threshold crossing in the continuous_data
        waveform_length - [before, after] number of datapoints to include in the waveform
    """
    # Using spike_indices create an array of indices (windows) to extract waveforms from LFP trace
    # The following values are chosen to match OpenEphysGUI default window
    spike_indices = np.expand_dims(spike_indices, 1)
    # Create windows for indexing all samples for a waveform
    windows = np.arange(waveform_length[0] + waveform_length[1], dtype=np.int64) - np.int64(waveform_length[0])
    windows = np.tile(windows, (spike_indices.size,1))
    windows = windows + np.tile(spike_indices, (1,windows.shape[1]))
    # Skip windows that are too close to edge of signal
    tooearly = windows < 0
    toolate = windows > (continuous_tetrode_data.shape[1] - 1)
    idx_delete = np.any(np.logical_or(tooearly, toolate), axis=1)
    windows = np.delete(windows, np.where(idx_delete)[0], axis=0)
    spike_indices = np.delete(spike_indices, np.where(idx_delete)[0], axis=0)
    # Prepare idx_keep array to return
    idx_keep = np.logical_not(idx_delete).squeeze()
    # Create indexing for channels and spikes
    # windows and windows_channels shape must be  nspikes x nchans x windowsize
    windows = np.repeat(windows[:,:,np.newaxis], 4, axis=2)
    windows = np.swapaxes(windows,1,2)
    windows_channels = np.arange(4, dtype=np.int64)
    windows_channels = np.tile(windows_channels[np.newaxis,:,np.newaxis], 
                               (windows.shape[0], 1, windows.shape[2]))
    waveforms = continuous_tetrode_data[windows_channels,windows]
    
    return waveforms, spike_indices, idx_keep

def filter_spike_data(spike_data_tet, pos_edges, threshold, noise_cut_off, verbose=True):
    """
    Filters data on all tetrodes according to multiple criteria
    Inputs:
        spike_data [list] - list of dictionaries with 'waveforms' and 'timestamps' for each tetrode
        pos_edges [list] - first and last position data timestamp
        threshold - threshold value in microvolts
        noise_cut_off - value for removing spikes above this cut off in microvolts
    """
    # Create idx_keep all True for all tetrodes. Then turn values to False with each filter
    idx_keep = np.ones(spike_data_tet['waveforms'].shape[0], dtype=np.bool)
    # Include spikes occured during position data
    idx = np.logical_and(spike_data_tet['timestamps'] > pos_edges[0], 
                         spike_data_tet['timestamps'] < pos_edges[1])
    idx = idx.squeeze()
    idx_keep = np.logical_and(idx_keep, idx)
    # Include spikes above threshold
    threshold_int16 = np.int16(np.round(threshold / 0.195))
    idx = np.any(np.any(spike_data_tet['waveforms'] < -threshold_int16, axis=2), axis=1)
    idx_keep = np.logical_and(idx_keep, idx)
    if verbose and np.sum(idx) < idx.size:
        percentage_above_threshold = np.sum(idx) / float(idx.size) * 100
        print('{:.1f}% of spikes on tetrode {} were above threshold'.format(percentage_above_threshold, spike_data_tet['nr_tetrode'] + 1))
    # Include spikes below noise cut off
    if noise_cut_off and (noise_cut_off != 0):
        noise_cut_off_int16 = np.int16(np.round(noise_cut_off / 0.195))
        idx = np.all(np.all(np.abs(spike_data_tet['waveforms']) < noise_cut_off_int16, axis=2), axis=1)
        idx_keep = np.logical_and(idx_keep, idx)
        if verbose and np.sum(idx) < idx.size:
            percentage_too_big = (1 - np.sum(idx) / float(idx.size)) * 100
            print('{:.1f}% of spikes removed on tetrode {}'.format(percentage_too_big, spike_data_tet['nr_tetrode'] + 1))
    idx_keep = idx_keep.reshape(idx_keep.size)

    return idx_keep

def clarify_OpenEphysDataPaths(OpenEphysDataPaths):
    # If directories entered as paths, attempt creating path to file by appending experiment_1.nwb
    for ndata, OpenEphysDataPath in enumerate(OpenEphysDataPaths):
        if not os.path.isfile(OpenEphysDataPath):
            new_path = os.path.join(OpenEphysDataPath, 'experiment_1.nwb')
            if os.path.isfile(new_path):
                OpenEphysDataPaths[ndata] = new_path
            else:
                raise ValueError('The following path does not lead to a NWB data file:\n' + OpenEphysDataPath)

    return OpenEphysDataPaths


def get_channel_map(OpenEphysDataPaths):
    # Get channel maps for all datasets
    channel_maps = []
    for OpenEphysDataPath in OpenEphysDataPaths:
        if NWBio.check_if_settings_available(OpenEphysDataPaths[0],'/General/channel_map/'):
            channel_maps.append(NWBio.load_settings(OpenEphysDataPaths[0],'/General/channel_map/'))
        else:
            raise Exception('No channel map for: ' + OpenEphysDataPath)
    # Ensure that channel map is the same in all datasets
    if not all([NWBio.check_if_channel_maps_are_same(channel_maps[0], x) for x in channel_maps]):
        raise Exception('Not all channel maps are the same in: ' + str(OpenEphysDataPaths))
    channel_map = channel_maps[0]
    # Check if the channel list fully coveres a set of tetrodes
    for area in channel_map.keys():
        if np.mod(len(channel_map[area]['list']), 4) != 0:
            raise ValueError('Channel map range must map to full tetrodes.')

    return channel_map


def process_position_data(OpenEphysDataPath, postprocess=False, maxjump=25):
    if NWBio.check_if_tracking_data_available(OpenEphysDataPath):
        print('Processing tracking data for: ' + OpenEphysDataPath)
        ProcessedPos = process_tracking_data(OpenEphysDataPath, verbose=True)
        NWBio.save_tracking_data(OpenEphysDataPath, ProcessedPos, ProcessedPos=True, overwrite=True)
        print('ProcessedPos saved to ' + OpenEphysDataPath)
    elif NWBio.check_if_binary_pos(OpenEphysDataPath):
        use_binary_pos(OpenEphysDataPath, postprocess=postprocess, maxjump=maxjump)
        print('Using binary position data for: ' + OpenEphysDataPath)
    else:
        print('Proceeding without position data for: ' + OpenEphysDataPath)


def ensure_processed_position_data_is_available(OpenEphysDataPath, **kwargs):
    if not NWBio.check_if_processed_position_data_available(OpenEphysDataPath):
        process_position_data(OpenEphysDataPath, **kwargs)


def ensure_data_available_for_all_tetrodes(spike_data, tetrode_nrs):
    # Check which tetrodes have data missing in the recording
    tetrodes_missing_in_spike_data = copy.deepcopy(tetrode_nrs)
    for data in spike_data:
        if 'waveforms' in data.keys():
            tetrodes_missing_in_spike_data.remove(data['nr_tetrode'])
    if len(tetrodes_missing_in_spike_data) > 0:
        raise Exception('No data for tetrodes: ' + ', '.join(map(str, tetrodes_missing_in_spike_data)))

def combine_spike_datas_tet(spike_datas_tet):
    # Combine all spike_datas into a single spike_data
    orig_spike_datas_tet = copy.deepcopy(spike_datas_tet)
    spike_datas_tet_comb = orig_spike_datas_tet[0]
    for tmp_spike_data in orig_spike_datas_tet[1:]:
        spike_datas_tet_comb['waveforms'] = np.append(spike_datas_tet_comb['waveforms'], 
                                                      tmp_spike_data['waveforms'], axis=0)
        spike_datas_tet_comb['timestamps'] = np.append(spike_datas_tet_comb['timestamps'], 
                                                       tmp_spike_data['timestamps'], axis=0)
        spike_datas_tet_comb['idx_keep'] = np.append(spike_datas_tet_comb['idx_keep'], 
                                                     tmp_spike_data['idx_keep'], axis=0)

    return spike_datas_tet_comb

def uncombine_spike_datas_tet_clusterIDs(clusterIDs, spike_datas_tet):
    clusterIDs = copy.copy(clusterIDs)
    # Extract clusters for each original spike_data
    for ndata in range(len(spike_datas_tet)):
        # Get clusterIDs for this dataset
        nspikes = sum(spike_datas_tet[ndata]['idx_keep'])
        spike_datas_tet[ndata]['clusterIDs'] = clusterIDs[list(range(nspikes))]
        # Remove these clusterIDs from the list
        clusterIDs = np.delete(clusterIDs, list(range(nspikes)), axis=0)

    return spike_datas_tet

def applyKlustaKwik_to_combined_recordings(spike_datas_tet, max_clusters=31):
    spike_datas_tet_comb = combine_spike_datas_tet(spike_datas_tet)
    clusterIDs = applyKlustaKwik_on_spike_data_tet(spike_datas_tet_comb, 
                                                   max_possible_clusters=max_clusters)
    spike_datas_tet = uncombine_spike_datas_tet_clusterIDs(clusterIDs, spike_datas_tet)

    return spike_datas_tet

def split_KiloSort_output(datas_comb_clusterIDs, datas_comb_spike_indices, datas_tet_shape):
    datas_clusterIDs = []
    datas_spike_indices = []
    data_start_pos = 0
    for data_shape in datas_tet_shape:
        nspikes = sum(datas_comb_spike_indices < data_start_pos + data_shape[1])
        datas_spike_indices.append(datas_comb_spike_indices[:nspikes] - data_start_pos)
        datas_clusterIDs.append(datas_comb_clusterIDs[:nspikes])
        datas_comb_spike_indices = np.delete(datas_comb_spike_indices, list(range(nspikes)))
        datas_comb_clusterIDs = np.delete(datas_comb_clusterIDs, list(range(nspikes)))
        data_start_pos += data_shape[1]

    return datas_clusterIDs, datas_spike_indices

def save_spike_data_to_disk(OpenEphysDataPath, processing_method, tetrode_nr, waveforms=None, timestamps=None, idx_keep=None, clusterIDs=None):
    spike_name = NWBio.get_spike_name_for_processing_method(processing_method)
    if not (waveforms is None) and not (timestamps is None):
        NWBio.save_spikes(OpenEphysDataPath, tetrode_nr, waveforms, 
                          timestamps, spike_name=spike_name, overwrite=True)
    if not (idx_keep is None):
        NWBio.save_tetrode_idx_keep(OpenEphysDataPath, tetrode_nr, idx_keep, 
                                    spike_name=spike_name, overwrite=True)
    if not (clusterIDs is None):
        NWBio.save_tetrode_clusterIDs(OpenEphysDataPath, tetrode_nr, clusterIDs, 
                                      spike_name=spike_name, overwrite=True)


class Multiprocess_KlustaKwik(object):

    def __init__(self):
        self.multiprocessor = hfunct.multiprocess()

    def add(self, spike_data_tet, max_clusters=31):
        self.multiprocessor.run(applyKlustaKwik_on_spike_data_tet, 
                                args=(spike_data_tet,), 
                                kwargs={'max_possible_clusters': max_clusters}, 
                                single_cpu_affinity=False)

    def get(self):
        return self.multiprocessor.results()


def process_combined_recordings_spike_datas(spike_datas, tetrode_nrs, max_clusters):
    mp_KlustaKwik = Multiprocess_KlustaKwik()
    for n_tet in range(len(tetrode_nrs)):
        hfunct.proceed_when_enough_memory_available(percent=0.60)
        spike_datas_tet = [spike_data[n_tet] for spike_data in spike_datas]
        spike_datas_tet_comb = combine_spike_datas_tet(spike_datas_tet)
        mp_KlustaKwik.add(spike_datas_tet_comb, max_clusters=max_clusters)
        hfunct.print_progress(n_tet + 1, len(tetrode_nrs), prefix='Applying KlustaKwik:', suffix=' T: ' + str(n_tet + 1) + '/' + str(len(tetrode_nrs)))
    for n_tet in range(len(tetrode_nrs)):
        # Get sorted combined data for one tetrode
        clusterIDs = mp_KlustaKwik.get()[n_tet]
        spike_datas_tet = [spike_data[n_tet] for spike_data in spike_datas]
        spike_datas_tet = uncombine_spike_datas_tet_clusterIDs(clusterIDs, spike_datas_tet)
        for n_dataset, spike_data_tet in enumerate(spike_datas_tet):
            spike_datas[n_dataset][n_tet] = spike_data_tet

    return spike_datas


def process_available_spikes_using_klustakwik(OpenEphysDataPaths, channels, 
                                              noise_cut_off=1000, threshold=50, 
                                              max_clusters=31):
    tetrode_nrs = hfunct.get_tetrode_nrs(channels)
    # Load spikes
    spike_datas = [list(range(len(tetrode_nrs))) for i in range(len(OpenEphysDataPaths))]
    for n_dataset, OpenEphysDataPath in enumerate(OpenEphysDataPaths):
        print('Loading data for processing: ' + OpenEphysDataPath)
        spike_data = NWBio.load_spikes(OpenEphysDataPath, tetrode_nrs=tetrode_nrs, use_badChan=True)
        ensure_data_available_for_all_tetrodes(spike_data, tetrode_nrs)
        # Find eligible spikes on all tetrodes
        pos_edges = NWBio.get_processed_tracking_data_timestamp_edges(OpenEphysDataPath)
        for spike_data_tet in spike_data:
            spike_data_tet['idx_keep'] = filter_spike_data(spike_data_tet, pos_edges, threshold, noise_cut_off)
        # Combine into spike_datas
        for n_tet, spike_data_tet in enumerate(spike_data): 
            spike_datas[n_dataset][n_tet] = spike_data_tet
    # Cluster each tetrode using KlustaKwik. This creates 'clusterIDs' field in spike_data dictionaries.
    if len(spike_datas) == 1:
        hfunct.print_progress(0, len(tetrode_nrs), prefix='Applying KlustaKwik:', suffix=' T: 0/' + str(len(tetrode_nrs)), initiation=True)
        mp_KlustaKwik = Multiprocess_KlustaKwik()
        for n_tet in range(len(tetrode_nrs)):
            mp_KlustaKwik.add(spike_datas[0][n_tet], max_clusters=max_clusters)
            hfunct.print_progress(n_tet + 1, len(tetrode_nrs), prefix='Applying KlustaKwik:', suffix=' T: ' + str(n_tet + 1) + '/' + str(len(tetrode_nrs)))
        for n_tet in range(len(tetrode_nrs)):
            spike_datas[0][n_tet]['clusterIDs'] = mp_KlustaKwik.get()[n_tet]
    elif len(spike_datas) > 1:
        spike_datas = process_combined_recordings_spike_datas(spike_datas, tetrode_nrs, max_clusters)
    # Overwrite clusterIDs on disk
    for OpenEphysDataPath, spike_data in zip(OpenEphysDataPaths, spike_datas):
        print('Saving processing output to: ' + OpenEphysDataPath)
        for spike_data_tet in spike_data:
            save_spike_data_to_disk(OpenEphysDataPath, 'klustakwik', spike_data_tet['nr_tetrode'], 
                                    idx_keep=spike_data_tet['idx_keep'], 
                                    clusterIDs=spike_data_tet['clusterIDs'])

    return spike_datas

def process_spikes_from_raw_data_using_klustakwik(OpenEphysDataPaths, channels, 
                                                  noise_cut_off=1000, threshold=50, 
                                                  max_clusters=31):
    tetrode_nrs = hfunct.get_tetrode_nrs(channels)
    tooclose = 30
    spike_datas = [list(range(len(tetrode_nrs))) for i in range(len(OpenEphysDataPaths))]
    # Preload continuous data
    preloaded_datas = []
    for OpenEphysDataPath in OpenEphysDataPaths:
        print('Loading data for processing: ' + OpenEphysDataPath)
        preloaded_data = ContinuousDataPreloader(OpenEphysDataPath, channels)
        preloaded_data.prepare_referencing('other_channels')
        preloaded_datas.append(preloaded_data)
    hfunct.print_progress(0, len(tetrode_nrs), prefix='Extract & KlustaKwik:', suffix=' T: 0/' + str(len(tetrode_nrs)), initiation=True)
    for n_tet, tetrode_nr in enumerate(tetrode_nrs):
        # Load this tetrode for all datasets
        spike_datas_tet = []
        for n_dataset in range(len(OpenEphysDataPaths)):
            data_tet = preloaded_datas[n_dataset].get_channels(hfunct.tetrode_channels(tetrode_nr), 
                                                              referenced=True, filter_freqs=[300, 6000])
            spike_indices = detect_threshold_crossings_on_tetrode(data_tet, threshold, tooclose, 
                                                                  detection_method='negative')
            waveforms, spike_indices, _ = extract_spikes_from_tetrode(data_tet, spike_indices, 
                                                                   waveform_length=[6, 34])
            # Arrange waveforms, timestamps and tetrode number into a dictionary
            timestamps = preloaded_datas[n_dataset].timestamps[spike_indices].squeeze()
            spike_data_tet = {'waveforms': np.int16(waveforms), 
                              'timestamps': np.float64(timestamps), 
                              'nr_tetrode': tetrode_nr}
            # Create idx_keep field for this tetrode
            pos_edges = NWBio.get_processed_tracking_data_timestamp_edges(OpenEphysDataPaths[n_dataset])
            spike_data_tet['idx_keep'] = filter_spike_data(spike_data_tet, pos_edges, 
                                                           threshold, noise_cut_off, verbose=False)
            spike_datas_tet.append(spike_data_tet)
        # Apply KlustaKwik to this tetrode
        if len(spike_datas_tet) == 1:
            clusterIDs = applyKlustaKwik_on_spike_data_tet(spike_datas_tet[0], 
                                                           max_possible_clusters=max_clusters)
            spike_datas_tet[0]['clusterIDs'] = clusterIDs
        elif len(spike_datas_tet) > 1:
            spike_datas_tet = applyKlustaKwik_to_combined_recordings(spike_datas_tet, 
                                                                     max_clusters=max_clusters)
        # Put this tetrode from all datasets into spike_datas
        for n_dataset, spike_data_tet in enumerate(spike_datas_tet):
                spike_datas[n_dataset][n_tet] = spike_data_tet
        hfunct.print_progress(n_tet + 1, len(tetrode_nrs), prefix='Extract & KlustaKwik:', suffix=' T: ' + str(n_tet + 1) + '/' + str(len(tetrode_nrs)))
    # Close pre-loaded datasets
    for preloaded_data in preloaded_datas:
        preloaded_data.close()
    # Save spike_datas to disk
    for OpenEphysDataPath, spike_data in zip(OpenEphysDataPaths, spike_datas):
        for data_tet in spike_data:
            save_spike_data_to_disk(OpenEphysDataPath, 'klustakwik_raw', data_tet['nr_tetrode'], 
                                    waveforms=data_tet['waveforms'], timestamps=data_tet['timestamps'], 
                                    idx_keep=data_tet['idx_keep'], clusterIDs=data_tet['clusterIDs'])

    return spike_datas

def process_raw_data_with_kilosort(OpenEphysDataPaths, channels, noise_cut_off=1000, threshold=5, 
                                   num_clusters=31):
    KiloSortBinaryFileName = 'experiment_1.dat'
    tetrode_nrs = hfunct.get_tetrode_nrs(channels)
    spike_datas = [list(range(len(tetrode_nrs))) for i in range(len(OpenEphysDataPaths))]
    # Preload continuous data
    preloaded_datas = []
    for OpenEphysDataPath in OpenEphysDataPaths:
        print('Loading data for processing: ' + OpenEphysDataPath)
        preloaded_data = ContinuousDataPreloader(OpenEphysDataPath, channels)
        preloaded_data.prepare_referencing('other_channels')
        preloaded_datas.append(preloaded_data)
    # Start matlab engine
    eng = matlab.engine.start_matlab()
    eng.cd(os.path.join(package_config.package_path, 'Utils', 'KiloSortScripts'))
    # Work through each tetrode
    for n_tet, tetrode_nr in enumerate(tetrode_nrs):
        print('Applying KiloSort to tetrode ' + str(n_tet + 1) + '/' + str(len(tetrode_nrs)))
        KiloSortProcessingFolder = tempfile.mkdtemp('KiloSortProcessing')
        # Load this tetrode for all datasets
        datas_tet_shape = []
        datas_tet = []
        for n_dataset in range(len(OpenEphysDataPaths)):
            data_tet = preloaded_datas[n_dataset].get_channels(hfunct.tetrode_channels(tetrode_nr), referenced=True, 
                                                               filter_freqs=False, no_badChan=True)
            datas_tet_shape.append(data_tet.shape)
            datas_tet.append(data_tet)
        # Save continuous data as a binary file
        datas_tet = np.concatenate(datas_tet, axis=1)
        datas_tet = np.transpose(datas_tet)
        datas_tet.tofile(os.path.join(KiloSortProcessingFolder, KiloSortBinaryFileName))
        del datas_tet
        # Run KiloSort
        eng.master_file(float(datas_tet_shape[0][0]), KiloSortProcessingFolder, 
                        float(num_clusters), nargout=0)
        eng.clear(nargout=0)
        # Load KiloSort output
        datas_comb_clusterIDs = np.load(os.path.join(KiloSortProcessingFolder, 'spike_clusters.npy'))[:,0]
        datas_comb_clusterIDs = np.int16(datas_comb_clusterIDs) + 1
        datas_comb_spike_indices = np.load(os.path.join(KiloSortProcessingFolder, 'spike_times.npy'))[:,0]
        datas_comb_spike_indices = np.int64(datas_comb_spike_indices)
        # Delete KiloSort Processing folder
        shutil.rmtree(KiloSortProcessingFolder)
        # Separate clusterIDs and spike indices to different datasets
        datas_clusterIDs, datas_spike_indices = split_KiloSort_output(datas_comb_clusterIDs, datas_comb_spike_indices, datas_tet_shape)
        # Create spike_data dictionary for this tetrode for each dataset
        for n_dataset in range(len(OpenEphysDataPaths)):
            data_tet = preloaded_datas[n_dataset].get_channels(hfunct.tetrode_channels(tetrode_nr), referenced=True, 
                                                               filter_freqs=[300, 6000], no_badChan=False)
            waveforms, spike_indices, idx_keep = extract_spikes_from_tetrode(data_tet, datas_spike_indices[n_dataset], 
                                                                             waveform_length=[6, 34])
            # Only keep clusterIDs for clusters that were included by extract_spikes_from_tetrode
            datas_clusterIDs[n_dataset] = (datas_clusterIDs[n_dataset][idx_keep]).squeeze()
            # Arrange waveforms, timestamps and tetrode number into a dictionary
            timestamps = preloaded_datas[n_dataset].timestamps[spike_indices].squeeze()
            spike_data_tet = {'waveforms': np.int16(waveforms), 
                              'timestamps': np.float64(timestamps), 
                              'clusterIDs': np.int16(datas_clusterIDs[n_dataset]), 
                              'nr_tetrode': tetrode_nr}
            # Create idx_keep field for this tetrode
            pos_edges = NWBio.get_processed_tracking_data_timestamp_edges(OpenEphysDataPaths[n_dataset])
            spike_data_tet['idx_keep'] = filter_spike_data(spike_data_tet, pos_edges, 
                                                           threshold, noise_cut_off, verbose=False)
            # Only keep clusterIDs based on idx_keep to conform to KlustaKwik processing format
            spike_data_tet['clusterIDs'] = spike_data_tet['clusterIDs'][spike_data_tet['idx_keep']]
            # Position spike_data_tet to the list of spike_data for each dataset
            spike_datas[n_dataset][n_tet] = spike_data_tet
    # Close pre-loaded datasets
    for preloaded_data in preloaded_datas:
        preloaded_data.close()
    # Save spike_datas to disk
    for OpenEphysDataPath, spike_data in zip(OpenEphysDataPaths, spike_datas):
        for data_tet in spike_data:
            save_spike_data_to_disk(OpenEphysDataPath, 'kilosort', data_tet['nr_tetrode'], 
                                    waveforms=data_tet['waveforms'], timestamps=data_tet['timestamps'], 
                                    idx_keep=data_tet['idx_keep'], clusterIDs=data_tet['clusterIDs'])

    return spike_datas


def processing(OpenEphysDataPaths, processing_method='klustakwik', channel_map=None, 
               noise_cut_off=1000, threshold=50, make_AxonaData=False, 
               axonaDataArgs=(None, False), max_clusters=31, 
               force_position_processing=False, pos_data_processing_kwargs={}):
    # Ensure correct format for data paths
    try: # Python 3 workaround
        basestring
    except NameError:
        basestring = str
    if isinstance(OpenEphysDataPaths, basestring):
        OpenEphysDataPaths = [OpenEphysDataPaths]
    OpenEphysDataPaths = clarify_OpenEphysDataPaths(OpenEphysDataPaths)
    # Create ProcessedPos if not yet available
    for OpenEphysDataPath in OpenEphysDataPaths:
        if force_position_processing:
            process_position_data(OpenEphysDataPath, **pos_data_processing_kwargs)
        else:
            ensure_processed_position_data_is_available(OpenEphysDataPath, **pos_data_processing_kwargs)
    # Get channel_map if not available
    if channel_map is None:
        channel_map = get_channel_map(OpenEphysDataPaths)
    # Process spikes using specified method
    area_spike_datas = []
    print(hfunct.time_string(), 'DEBUG: Starting Processing', processing_method)
    DEBUG_Time = time()
    for area in channel_map.keys():
        channels = channel_map[area]['list']
        if processing_method == 'klustakwik':
            area_spike_datas.append(process_available_spikes_using_klustakwik(OpenEphysDataPaths, channels, 
                                                                              noise_cut_off=noise_cut_off, 
                                                                              threshold=threshold, 
                                                                              max_clusters=max_clusters))
        elif processing_method == 'klustakwik_raw':
            area_spike_datas.append(process_spikes_from_raw_data_using_klustakwik(OpenEphysDataPaths, channels, 
                                                                                  noise_cut_off=noise_cut_off, 
                                                                                  threshold=threshold, 
                                                                                  max_clusters=max_clusters))
        elif processing_method == 'kilosort':
            if not matlab_available:
                raise Exception('Matlab not available. Can not process using KiloSort.')
            area_spike_datas.append(process_raw_data_with_kilosort(OpenEphysDataPaths, channels, 
                                                                   noise_cut_off=noise_cut_off, threshold=5, 
                                                                   num_clusters=max_clusters))
    print(hfunct.time_string(), 'DEBUG: Finished Processing in ', time() - DEBUG_Time)
    # Save data in Axona Format
    del area_spike_datas
    if make_AxonaData:
        spike_name = NWBio.get_spike_name_for_processing_method(processing_method)
        for OpenEphysDataPath in OpenEphysDataPaths:
            createAxonaData_for_NWBfile(OpenEphysDataPath, spike_name=spike_name, 
                                        channel_map=channel_map, pixels_per_metre=axonaDataArgs[0], 
                                        show_output=axonaDataArgs[1])


def process_data_tree(root_path, only_keep_processor=None, downsample=False, delete_raw=False, max_clusters=31):
    # Create list of dirnames skipped
    dir_names_skipped = []
    # Commence directory walk
    for dir_name, subdirList, fileList in os.walk(root_path):
        for fname in fileList:
            if 'Experiment' in dir_name:
                if not (dir_name in dir_names_skipped):
                    dir_names_skipped.append(dir_name)
                    raise Warning('Experiment found in directory name, skipping: ' + dir_name)
            else:
                fpath = os.path.join(dir_name, fname)
                if fname == 'experiment_1.nwb':
                    AxonaDataExists = any(['AxonaData' in subdir for subdir in subdirList])
                    if not AxonaDataExists:
                        print(hfunct.time_string() + ' Applying KlustaKwik on ' + fpath)
                        processing(fpath, processing_method='klustakwik',
                                   noise_cut_off=1000, threshold=50, make_AxonaData=True,
                                   axonaDataArgs=(None, False), max_clusters=max_clusters)
                    if not (only_keep_processor is None):
                        # Get all processor paths in this file
                        processor_paths = NWBio.get_all_processor_paths(fpath)
                        # Check that the processor key requested to keep is available
                        if not any([path.endswith(only_keep_processor) for path in processor_paths]):
                            raise ValueError('{} not found in {}. Can not delete others.'.format(only_keep_processor,
                                                                                                 fpath))
                        # Delete all other processor paths in the file, if any available
                        if len(processor_paths) > 1:
                            for path in processor_paths:
                                if not path.endswith(only_keep_processor):
                                    print('Deleting path: {}\n    in file: {}'.format(path, fpath))
                                    NWBio.delete_path_in_file(fpath, path)
                    if downsample:
                        if not NWBio.check_if_downsampled_data_available(fpath):
                            if NWBio.check_if_raw_data_available(fpath):
                                print(hfunct.time_string() + ' Downsample ' + fpath)
                                create_downsampled_data(fpath, n_tetrodes=32, downsample_factor=10)
                            else:
                                print('Warning', 'Neither Downsampled or Raw data is available, skipping ' + fpath)
                        else:
                            print('Warning', 'Downsampled data already available, skipping ' + fpath)
                    if delete_raw:
                        if NWBio.check_if_raw_data_available(fpath):
                            print(hfunct.time_string() + ' Repack ' + fpath)
                            delete_raw_data(fpath, only_if_downsampled_data_available=True)
                        else:
                            print('Warning', 'No raw data to be deleted in ' + fpath)


def main()
    # Input argument handling and help info
    parser = argparse.ArgumentParser(description='Apply KlustaKwik and export into Axona format.')
    parser.add_argument('paths', type=str, nargs='*', 
                        help='recording data folder(s) (can enter multiple paths separated by spaces to KlustaKwik simultaneously)')
    parser.add_argument('--chan', type=int, nargs = 2, 
                        help='list the first and last channel to process (counting starts from 1)')
    parser.add_argument('--noisecut', type=int, nargs = 1, 
                        help='enter 0 to skip or value in microvolts for noise cutoff (default is 1000)')
    parser.add_argument('--threshold', type=int, nargs = 1, 
                        help='enter spike threshold in microvolts (default is 50)')
    parser.add_argument('--klustakwik', action='store_true',
                        help='use KlustaKwik to cluster spike data saved by OpenEphys GUI (default)')
    parser.add_argument('--klustakwik_raw', action='store_true',
                        help='use KlustaKwik to cluster spike data obtained from raw data')
    parser.add_argument('--kilosort', action='store_true',
                        help='use KiloSort to cluster spike data (this uses raw data only)')
    parser.add_argument('--noAxonaData', action='store_true',
                        help='to skip conversion into AxonaData format after processing')
    parser.add_argument('--ppm', type=int, nargs = 1, 
                        help='(for AxonaData) enter pixels_per_metre to assume position data is in pixels')
    parser.add_argument('--force_position_processing', action='store_true',
                        help='instruct reprocessing position data and overwrite previous data')
    parser.add_argument('--position_postprocessing', action='store_true',
                        help='instruct position data postprocessing. Only applicable to binary position data.\n'
                             + '(default is no postprocessing)')
    parser.add_argument('--reprocess_tracking_in_directory', action='store_true',
                        help=('Re-process tracking data with default settings in all.\n'
                              + 'files named experiment_1.nwb in all sub-folders from input directory.'))
    parser.add_argument('--position_maxjump', type=float, nargs = 1, 
                        help='enter maximum allowed jump in position data values for position data postprocessing')
    parser.add_argument('--max_clusters', type=int, nargs = 1, 
                        help='Specifies the maximum number of cluster to find. Default is 31.')
    parser.add_argument('--show_output', action='store_true', 
                        help='(for AxonaData) to open AxonaData output folder after processing')
    parser.add_argument('--datatree', action='store_true', 
                        help='to process a whole data tree with default arguments in method process_data_tree')
    parser.add_argument('--only_keep_processor', type=str, nargs=1,
                        help=('All other continuous data subsets are deleted. Specify which one to keep.\n'
                              + 'Only available with datatree option.'))
    parser.add_argument('--downsample', action='store_true', 
                        help='to downsample a whole data tree after processing. Only available with datatree option.')
    parser.add_argument('--delete_raw', action='store_true',
                        help='to delete raw data after downsampling. Only available with datatree and downsample option.')
    parser.add_argument('--verbose', action='store_true',
                        help='Verbosity of progress and warnings. Default is False (off).')
    args = parser.parse_args()
    # Get paths to recording files
    OpenEphysDataPaths = args.paths
    # If reprocessing tracking in directory is requested, just do that
    if args.reprocess_tracking_in_directory:
        if len(OpenEphysDataPaths) > 1:
            raise Exception('Only one root path should be specified if reprocess_tracking_in_directory is set.')
        recompute_tracking_data_for_all_files_in_directory_tree(OpenEphysDataPaths[0],
                                                                verbose=args.verbose[0])
    # If datatree processing requested, use process_data_tree method
    if args.datatree:
        if args.only_keep_processor:
            only_keep_processor = args.only_keep_processor[0]
        else:
            only_keep_processor = None
        if args.downsample:
            downsample = True
        else:
            downsample = False
        if args.delete_raw:
            delete_raw = True
        else:
            delete_raw = False
        process_data_tree(OpenEphysDataPaths[0], only_keep_processor, downsample, delete_raw)
    else:
        # Get chan input variable
        if args.chan:
            chan = [args.chan[0] - 1, args.chan[1]]
            if np.mod(chan[1] - chan[0], 4) != 0:
                raise ValueError('Channel range must cover full tetrodes')
            area_name = 'Chan' + str(args.chan[0]) + '-' + str(args.chan[1])
            channel_map = {area_name: {'list': list(range(chan[0], chan[1], 1))}}
        else:
            channel_map = None
        # Rewrite default noisecut if specified
        if args.noisecut:
            noise_cut_off = args.noisecut[0]
        else:
            noise_cut_off = 1000
        # Rewrite default threshold if specified
        if args.threshold:
            threshold = args.threshold[0]
        else:
            threshold = 50
        # Specify processing_method
        if args.klustakwik:
            processing_method = 'klustakwik'
        elif args.klustakwik_raw:
            processing_method = 'klustakwik_raw'
        elif args.kilosort:
            processing_method = 'kilosort'
        else:
            processing_method = 'klustakwik'
        # Specify position data processing options
        if args.force_position_processing:
            force_position_processing = True
        else:
            force_position_processing = False
        pos_data_processing_kwargs = {}
        if args.position_postprocessing:
            pos_data_processing_kwargs['postprocess'] = True
        else:
            pos_data_processing_kwargs['postprocess'] = False
        if args.position_maxjump:
            pos_data_processing_kwargs['maxjump'] = args.position_maxjump[0]
        # Specify axona data options
        if args.noAxonaData:
            make_AxonaData = False
        else:
            make_AxonaData = True
        if args.ppm:
            pixels_per_metre = args.ppm[0]
        else:
            pixels_per_metre = None
        if args.max_clusters:
            max_clusters = args.max_clusters[0]
        else:
            max_clusters = 31
        if args.show_output:
            show_output = True
        else:
            show_output = False
        axonaDataArgs = (pixels_per_metre, show_output)
        # Run the script
        processing(OpenEphysDataPaths, processing_method, channel_map, noise_cut_off, 
                   threshold, make_AxonaData, axonaDataArgs, max_clusters, 
                   force_position_processing, pos_data_processing_kwargs)


if __name__ == '__main__':
    main()



import argparse
import piconzero as pz
from time import sleep
from ZMQcomms import paired_messenger
from threading import Thread
import os
from audioSignalGenerator import createAudioSignal

import contextlib
with contextlib.redirect_stdout(None):
    import pygame

class piconzero_state_controller(object):
    '''
    This class is used to control the state of piconzero
    where multiple processes may use it interchangably
    '''

    def __init__(self):
        self.piconzero_initialized = False

    def ensure_piconzero_is_initialized(self):
        if not self.piconzero_initialized:
            pz.init()
            self.piconzero_initialized = True

    def ensure_piconzero_is_closed(self):
        if self.piconzero_initialized:
            pz.cleanup()
            self.piconzero_initialized = False

# Instantiate piconzero_state_controller to be used by any function in this script
pz_state_controller = piconzero_state_controller()


class pinchValveController(object):
    '''
    This controls the current driven to pinch valve via Picon Zero
    '''
    def __init__(self, motor_pin=1, max_power=127):
        pz_state_controller.ensure_piconzero_is_initialized()
        self.motor_pin = motor_pin
        self.max_power = max_power

    def openTimer(self, duration):
        '''
        Fully opens pinch valve for specified time in seconds
        '''
        pz.setMotor(self.motor_pin, self.max_power)
        sleep(duration)
        pz.setMotor(self.motor_pin, 0)

    def close(self):
        pz_state_controller.ensure_piconzero_is_closed()


class LEDcontroller(object):
    '''
    This controls an LED light via Picon Zero.
    '''
    def __init__(self, intensity=100, pin=0):
        '''
        Itensity (0 - 100) of LED brightness and its pin on Picon Zero can be specified.
        '''
        pz_state_controller.ensure_piconzero_is_initialized()
        self.intensity = int(intensity)
        self.pin = pin
        if self.intensity == 100:
            pz.setOutputConfig(self.pin, 0)
        else:
            pz.setOutputConfig(self.pin, 1)

    def turnLEDon(self):
        pz.setOutput(self.pin, self.intensity)

    def turnLEDoff(self):
        pz.setOutput(self.pin, 0)

    def close(self):
        pz_state_controller.ensure_piconzero_is_closed()


class audioSignalController(object):
    '''
    This class creates a signal with specific parameters and can play it from a local speaker.
    '''
    def __init__(self):
        '''
        The sound frequency band center, width and modulating frequency can be defiend in Hz
        '''
        # Initialize pygame for playing sound
        os.system('amixer sset PCM,0 99%')
        pygame.mixer.pre_init(48000, -16, 2)
        pygame.mixer.init()
        pygame.init()
        # Create sound
        self.sound = {}

    def createAudioSignal(self, name, frequency, frequency_band_width, modulation_frequency, loudness='default'):
        if loudness == 'loud':
            self.sound[name] = createAudioSignal(frequency, frequency_band_width, modulation_frequency, 200000)
        else:
            self.sound[name] = createAudioSignal(frequency, frequency_band_width, modulation_frequency)

    def playAudioSignal(self, name):
        self.sound[name].play(-1)

    def stopAudioSignal(self, name):
        self.sound[name].stop()

    def playAudioSignalForSpecificDuration(self, name, duration):
        self.sound[name].play(-1, maxtime=int(duration * 1000))

    def startAllSounds(self):
        for key in self.sound.keys():
            self.sound[key].play(-1)

    def stopAllSounds(self):
        for key in self.sound.keys():
            self.sound[key].stop()

    def close(self):
        self.stopAllSounds()
        pygame.mixer.quit()
        pygame.quit()


class Controller(object):
    '''
    This class controls the operation of the pinch valve, speaker and an LED on the milkFeeder.
    '''

    def __init__(self, pinchValve=False, trialAudioSignalParams=None, lightSignalIntensity=None, 
                 lightSignalPins=[0], negativeAudioSignal=None, init_feedback=False):
        '''
        pinchValve             - bool - makes pinch valve release available
        trialAudioSignalParams - tuple(signal_frequency,frequency_band_width,modulation_frequency) -
                                 Initializes audioSignal controller with specified parameters.
        lightSignalIntensity   - int - Initializes lightSignal controller with specified intensity (0 - 100).
        lightSignalPins        - list of int - number for picon zero pins to use for LED signalling.
        negativeAudioSignal    - float - allows playing a full-range white noise for specified duration if > 0
        init_feedback          - bool - Sends ZMQ message to localhost if initialization successful: init_successful
                                 When using this, ensure a ZMQ device is paired to the IP of the device running this script.
        '''
        self.pinchValve = bool(pinchValve)
        # Initialize ZMQcomms in a separate thread
        T_initialize_ZMQcomms = Thread(target=self.initialize_ZMQcomms)
        T_initialize_ZMQcomms.start()
        # Initialize all requested features
        if self.pinchValve:
            self.pinchValveController = pinchValveController()
        if not (lightSignalIntensity is None) and lightSignalIntensity > 0:
            self.LEDcontrollers = []
            for pin in lightSignalPins:
                self.LEDcontrollers.append(LEDcontroller(lightSignalIntensity, pin))
        if not (trialAudioSignalParams is None) or not (negativeAudioSignal is None):
            self.audioSignalController = audioSignalController()
            if not (trialAudioSignalParams is None):
                self.audioSignalController.createAudioSignal('trialAudioSignal', *trialAudioSignalParams)
            if not (negativeAudioSignal is None):
                self.audioSignalController.createAudioSignal('negativeAudioSignal', 12000, 10000, 0, 'loud')
                self.negativeAudioSignalDuration = negativeAudioSignal
        # Ensure ZMQ comms have been initialized
        T_initialize_ZMQcomms.join()
        # Send feedback text if requested
        if init_feedback:
            self.ZMQmessenger.sendMessage('init_successful')

    def initialize_ZMQcomms(self):
        '''
        Allows initializing ZMQ comms in a separate thread
        '''
        self.ZMQmessenger = paired_messenger(port=4186)
        self.ZMQmessenger.add_callback(self.command_parser)
        sleep(1) # This ensures all ZMQ protocols have been properly initated before finishing this process

    def command_parser(self, message):
        '''
        Parses incoming ZMQ message for function name, input arguments and calls that function.
        '''
        method_name = message.split(' ')[0]
        args = message.split(' ')[1:]
        getattr(self, method_name)(*args)

    def releaseMilk(self, duration, feedback=False):
        '''
        Fully opens pinch valve for specified time in seconds. Input duration is converted to float.
        '''
        if type(duration) != float:
            duration = float(duration)
        if self.pinchValve:
            self.pinchValveController.openTimer(duration)
            if feedback:
                self.ZMQmessenger.sendMessage('releaseMilk successful')

    def startLightSignal(self):
        if hasattr(self, 'LEDcontrollers'):
            for LEDcontroller in self.LEDcontrollers:
                LEDcontroller.turnLEDon()

    def stopLightSignal(self):
        if hasattr(self, 'LEDcontrollers'):
            for LEDcontroller in self.LEDcontrollers:
                LEDcontroller.turnLEDoff()

    def startTrialAudioSignal(self):
        self.audioSignalController.playAudioSignal('trialAudioSignal')

    def stopTrialAudioSignal(self):
        self.audioSignalController.stopAudioSignal('trialAudioSignal')

    def startNegativeAudioSignal(self):
        self.audioSignalController.playAudioSignal('negativeAudioSignal')

    def playNegativeAudioSignal(self):
        self.audioSignalController.playAudioSignalForSpecificDuration('negativeAudioSignal', self.negativeAudioSignalDuration)

    def stopNegativeAudioSignal(self):
        self.audioSignalController.stopAudioSignal('negativeAudioSignal')

    def startAllSignals(self):
        if hasattr(self, 'LEDcontrollers'):
            self.startLightSignal()
        if hasattr(self, 'audioSignalController'):
            self.audioSignalController.startAllSounds()

    def stopAllSignals(self):
        if hasattr(self, 'LEDcontrollers'):
            self.stopLightSignal()
        if hasattr(self, 'audioSignalController'):
            self.audioSignalController.stopAllSounds()

    def keepProcessRunning(self):
        '''
        This method allows any process to wait until close method has been called for this class
        '''
        self.closeCommandReceived = False
        while not self.closeCommandReceived:
            try:
                sleep(0.1)
            except KeyboardInterrupt:
                self.close()

    def close(self):
        '''
        Closes all initiated processes gracefully.
        '''
        self.closeCommandReceived = True
        self.stopAllSignals()
        self.ZMQmessenger.close()
        if hasattr(self, 'pinchValveController'):
            self.pinchValveController.close()
        if hasattr(self, 'LEDcontrollers'):
            for LEDcontroller in self.LEDcontrollers:
                LEDcontroller.close()
        if hasattr(self, 'audioSignalController'):
            self.audioSignalController.close()


if __name__ == '__main__':
    # Input argument handling and help info
    parser = argparse.ArgumentParser(description='Running this script initates Controller class.')
    parser.add_argument('--pinchValve', action='store_true', 
                        help='Initializes pinch valve.')
    parser.add_argument('--lightSignalIntensity', type=int, nargs = 1, 
                        help='Initializes lightSignal controller with specified intensity (0 - 100).')
    parser.add_argument('--lightSignalPins', type=int, nargs='*', 
                        help='Specifies which pins to use for light signal.')
    parser.add_argument('--trialAudioSignal', type=int, nargs = 3, 
                        help='Initializes audioSignalController with specified parameters\n' + \
                        'signal_frequency frequency_band_width modulation_frequency.')
    parser.add_argument('--negativeAudioSignal', type=float, nargs = 1, 
                        help='Initializes audioSignalController also for wide-band noise burst of specified duration(s).')
    parser.add_argument('--init_feedback', action='store_true', 
                        help='Sends ZMQ message to localhost if initialization successful: init_successful\n' + \
                        'When using this, ensure a ZMQ device is paired to the IP of the device running this script.')
    parser.add_argument('--openValve', type=float, nargs = 1, 
                        help='Makes device opens pinch valve for specified duration (s) and does not activate Controller class.')
    args = parser.parse_args()
    # If releasePellet command given skip everything and just release the number of pellets specified
    if args.openValve:
        pvc = pinchValveController()
        pvc.openTimer(float(args.openValve[0]))
        pvc.close()
    else:
        # Load input arguments
        if args.lightSignalIntensity:
            pinchValve = True
        else:
            pinchValve = False
        if args.lightSignalIntensity:
            lightSignalIntensity = int(args.lightSignalIntensity[0])
        else:
            lightSignalIntensity = None
        if args.lightSignalPins:
            lightSignalPins = args.lightSignalPins
        else:
            lightSignalPins = [1]
        if args.trialAudioSignal:
            trialAudioSignalParams = (args.trialAudioSignal[0], args.trialAudioSignal[1], args.trialAudioSignal[2])
        else:
            trialAudioSignalParams = None
        if args.negativeAudioSignal:
            negativeAudioSignal = float(args.negativeAudioSignal[0])
            if negativeAudioSignal == 0:
                negativeAudioSignal = None
        else:
            negativeAudioSignal = None
        if args.init_feedback:
            init_feedback = True
        else:
            init_feedback = False
        # Initialize Controller
        active_Controller = Controller(pinchValve=pinchValve, trialAudioSignalParams=trialAudioSignalParams, 
                                       lightSignalIntensity=lightSignalIntensity, lightSignalPins=lightSignalPins, 
                                       negativeAudioSignal=negativeAudioSignal, init_feedback=init_feedback)
        # Start an endless loop that can be cancelled with close method or ctrl+c
        active_Controller.keepProcessRunning()



import zmq
from threading import Lock, Thread
from time import sleep, time
import traceback
import socket
import pickle


def get_localhost_ip():
    # Get local IP address
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    address = s.getsockname()[0]
    s.close()

    return address


class paired_messenger(object):
    """
    This class can exchange messages with another pair_messenger instance running on the same
        or different machine, pointed to the same IP address.
    Bind one of the instances to address='localhost' 
        and the other pointing to the corresponding IP address.
    Use add_callback method to add callbacks which will be called when a message
        is received. You can couple arguments with the callback if presented as a list
        or tuple, with callback being the first element.
    NOTE! On slower systems, the initialization process may take some time.
        If messages are expected to be received or sent immediately, a delay may be necessary
        after instantiation of this object. Depending on system, 0.25 to 1.0 seconds.
    """
    def __init__(self, address='localhost', port=5884, timeout=0.5, printMessages=False):

        # Identify if client or server instance
        if address == 'localhost':
            self.localhost = True
            address = get_localhost_ip()
        else:
            self.localhost = False

        # Connect to an address
        self.url = "tcp://%s:%d" % (address, port)
        context = zmq.Context()
        self.socket = context.socket(zmq.PAIR)
        if self.localhost:
            self.socket.bind(self.url)
        else:
            self.socket.connect(self.url)
        self.socket.RCVTIMEO = int(timeout * 1000)  # in milliseconds

        # Set callbacks list
        self.callbacks = []
        
        if printMessages:
            self.add_callback(lambda msg: print(msg))

        # Add verification callback
        self.verification_dict = {}
        self.add_callback(self._verification_check)

        # Wait a moment to allow connection to be established
        sleep(2)
        
        # Start listening thread
        self.lock = Lock()
        self.is_running = True
        self.thread = Thread(target=self._run)
        self.thread.start()

    def sendMessage(self, message, verify=False):
        """Sends message to the paired messenger.

        :param bytes message: sent to paired device
        :param bool verify: if True, sendMessage() waits until such message is received back
        """
        if verify:
            self.verification_dict[message] = False
        self.socket.send(message)
        if verify:
            while not self.verification_dict[message]:
                sleep(0.05)
            del self.verification_dict[message]

    def close(self):

        if self.thread.is_alive():

            self.lock.acquire()
            self.is_running = False
            self.lock.release()

            self.thread.join()

            self.socket.disconnect(self.url)

    def add_callback(self, cb):
        """
        The callback function is started in a new thread if a message is received.
        To pass arguments with callback function, input a list or a tuple with
            callback function as the first element and the rest as individual arguments.
        The callback function should be expecting a message input as a string. If additional
            arguments were passed using the list or tuple method, these arguments should be
            expected by the function in the same order after the message string.
        """
        self.callbacks.append(cb)

    def _send_message_to_callbacks(self, msg):

        for cb in self.callbacks:
            if isinstance(cb, list) or isinstance(cb, tuple):
                Thread(target=cb[0], args=(msg,) + tuple(cb[1:])).start()
            else:
                Thread(target=cb, args=(msg,)).start()

    def _process_message(self, msg):
        """
        Called for each received message.
        Execution blocks reception of new messages. Threading recommended.
        """
        self._send_message_to_callbacks(msg)

    def _run(self):

        while True:

            self.lock.acquire()
            running = self.is_running
            self.lock.release()

            if not running:
                break

            try:
                msg = self.socket.recv()
                self._process_message(msg)

            except zmq.ZMQError:
                pass

            sleep(.01)

    def _verification_check(self, msg):
        if msg in list(self.verification_dict.keys()):
            self.verification_dict[msg] = True


def decode_pickled_message(msg):
    return pickle.loads(msg)


def encode_pickled_message(data):
    return pickle.dumps(data)


class remote_controlled_object(paired_messenger):
    """
    When instantiated with an object this class executes any incoming commands on that object.
    The incoming commands are expected to be sent using remote_object_controller.

    Note! remote_object_controller must be instantiated pair() method called before
    remote_controlled_object is instantiated.

    This class and the object must remain in scope of an active process.

    Sending a 'close' command will call close command on the object,
    return value if requested and then closes the remote_controlled_object instance.
    """
    def __init__(self, obj, *args, **kwargs):
        """
        remote_controlled_object must be instantiated with the object as first input argument.

        See paired_messenger for other input arguments.
        """
        self.obj = obj
        super(remote_controlled_object, self).__init__(*args, **kwargs)
        self.sendMessage('handshake'.encode())

    @staticmethod
    def _parse_message(msg):
        # Extract command to call from message string
        command, msg = msg.split(' '.encode(), 1)
        command = command.decode()
        # Extract return_value request from remaining message string
        return_value, msg = msg.split(' '.encode(), 1)
        if return_value == 'True'.encode():
            return_value = True
        elif return_value == 'False'.encode():
            return_value = False
        else:
            raise ValueError('return_value was not as expected.')
        # Extract input arguments from remaining message string
        input_arguments = decode_pickled_message(msg)
        args = input_arguments['args']
        kwargs = input_arguments['kwargs']

        return command, return_value, args, kwargs

    def _process_command(self, msg):
        # Parse raw message
        command, return_value, args, kwargs = remote_controlled_object._parse_message(msg)
        # Execute command with or without return value
        if return_value:
            self._execute_command_with_return(command, args, kwargs)
        else:
            self._execute_command(command, args, kwargs)
        # If close command sent, this remote_controlled_object is also closed.
        if command == 'close':
            self.close()

    def _execute_command(self, command, args, kwargs):
        """
        Parses incoming ZMQ message for function name, input arguments and calls that function.
        """
        getattr(self.obj, command)(*args, **kwargs)
    
    def _execute_command_with_return(self, command, args, kwargs):
        """
        Parses incoming ZMQ message for function name, input arguments and calls that function.
        """
        return_value = getattr(self.obj, command)(*args, **kwargs)
        encoded_return_value = encode_pickled_message(return_value)
        self.sendMessage(encoded_return_value)

    def _process_message(self, msg):
        Thread(target=self._process_command, args=(msg,)).start()


class remote_controlled_class(remote_controlled_object):
    """
    Allows using a class with remote_controlled_object before it is instantiated.

    Note! remote_object_controller must be instantiated pair() method called before
    remote_controlled_class is instantiated.

    remote_controlled_class must remain in scope to function. This can be achieved by calling
    it with block=True (see __init__() method) or by regularly checking isAlive() method
    to see if 'close' command has been received.

    On a paired remote_object_controller instance the sendInitCommand method
    must be called to instantiate the class before sendCommand method can be used.

    Once the class has been instantiated, remote_controlled_class behaves as remote_controlled_object.
    """
    def __init__(self, C, block, *args, **kwargs):
        """
        C - class to be used
        block - bool - if True, remote_controlled_class blocks until 'close' command is received.

        See paired_messenger for other input arguments.
        """
        self.C = C
        self.class_instantiated = False
        self.keep_class_alive = True
        super(remote_controlled_class, self).__init__(None, *args, **kwargs)
        if block:
            while self.keep_class_alive:
                sleep(0.1)

    def _init_object(self, args, kwargs):
        self.obj = self.C(*args, **kwargs)
        self.class_instantiated = True
        self.sendMessage('init_confirmation'.encode())

    def _process_command(self, msg):
        if self.class_instantiated:
            super(remote_controlled_class, self)._process_command(msg)
        else:
            command, return_value, args, kwargs = self._parse_message(msg)
            if command == '__init__':
                self._init_object(args, kwargs)

    def isAlive(self):
        """
        Returns boolean whether the remote_controlled_class is still alive.
        Returns False after close command has been received and instantiated class has been closed.
        """
        return self.keep_class_alive

    def close(self, *args, **kwargs):
        super(remote_controlled_class, self).close(*args, **kwargs)
        self.keep_class_alive = False


class remote_object_controller(paired_messenger):
    """
    Provides access to an object on a remote device via ZMQ.

    Can be paired with either remote_controlled_object or remote_controlled_class.

    Note! remote_object_controller must be instantiated and pair() method called
    before remote_controlled_object or remote_controlled_class is instantiated.

    If paired with remote_controlled_class, sendInitCommand must be called once paired
    to use other methods.

    See sendCommand() method for how to send commands and receive return values.
    """
    def __init__(self, *args, **kwargs):
        """
        See paired_messenger for other input arguments.
        """
        self.wait_for_handshake = True
        self.new_return_message = False
        self.wait_for_init_confirmation = False
        super(remote_object_controller, self).__init__(*args, **kwargs)

    def pair(self, timeout=0):
        """
        Returns True if successfully paired. Returns False if timeout reached.

        timeout - float - in seconds. If timeout=0 (default), pair() waits indefinitely.
        """
        wait_start_time = time()
        while self.wait_for_handshake:
            sleep(0.1)
            if timeout > 0 and (time() - wait_start_time) > timeout:
                break
        
        return not self.wait_for_handshake

    @staticmethod
    def encode_input_arguments(args, kwargs):
        input_arguments = {'args': args, 
                           'kwargs': kwargs}
        return encode_pickled_message(input_arguments)

    def sendInitCommand(self, timeout, *args, **kwargs):
        """
        Returns True if class successfully instantiated. Returns False if timeout reached.

        timeout - float - in seconds. If timeout=0 (default), sendInitCommand() waits indefinitely.

        Any following arguments are passed into the class __init__().
        """
        self.wait_for_init_confirmation = True
        encoded_input_arguments = remote_object_controller.encode_input_arguments(args, kwargs)
        self.sendMessage('__init__'.encode() + ' '.encode()
                         + 'True'.encode() + ' '.encode()
                         + encoded_input_arguments)
        wait_start_time = time()
        while self.wait_for_init_confirmation:
            sleep(0.1)
            if 0 < timeout < (time() - wait_start_time):
                break
        
        return not self.wait_for_init_confirmation

    def sendCommand(self, command, return_value, *args, **kwargs):
        """
        command - str - Name of the command to call on the object controlled via ZMQ
        return_value - bool - Whether to return value from command call.
                              return_value=True blocks until return value is received.
        Any additional input arguments are used as input arguments in command call on controlled object.
        These additional input arguments are pickled, compressed on this end
        and uncompressed, unpickled on the paired device.
        """
        encoded_input_arguments = remote_object_controller.encode_input_arguments(args, kwargs)
        self.sendMessage(command.encode() + ' '.encode()
                         + str(return_value).encode() + ' '.encode()
                         + encoded_input_arguments)
        if return_value:
            return self._wait_for_return_value()

    def _wait_for_return_value(self):
        """
        Blocks until return value is available.
        Returns the return value.
        """
        while not self.new_return_message:
            sleep(0.1)
        return_value = decode_pickled_message(self.return_message)
        self.new_return_message = False
        
        return return_value

    def _process_return_message(self, msg):
        assert not self.new_return_message  # Assuming return values are always processed before next arrives.
        self.return_message = msg
        self.new_return_message = True

    def _process_message(self, msg):
        if self.wait_for_handshake:
            if msg == 'handshake'.encode():
                self.wait_for_handshake = False
        elif self.wait_for_init_confirmation:
            if msg == 'init_confirmation'.encode():
                self.wait_for_init_confirmation = False
        else:
            Thread(target=self._process_return_message, args=(msg,)).start()


class PublishToOpenEphys(object):
    """
    This class allows sending messages to Open Ephys GUI over ZMQ.
    When created with defulat inputs, it will connect to Open Ephys GUI.
    Use sendMessage method to send messages to Open Ephys GUI
    """
    def __init__(self, address='localhost', port=5556):
        # Set up ZMQ connection to OpenEphysGUI
        url = "tcp://%s:%d" % (address, port)
        context = zmq.Context()
        self.socket = context.socket(zmq.REQ)
        self.socket.connect(url)

    def sendMessage(self, message):
        """Encodes message string into bytes and sends to OpenEphysGUI

        :param str message:
        :return:
        """
        self.socket.send(message.encode())
        _ = self.socket.recv()

    def close(self):
        self.socket.close()


def SendOpenEphysSingleMessage(message):
    """
    This function creates ZMQ connection with Open Ephys GUI just to send one message.
    This is sufficiently fast for single messages that are not very time sensitive.
    """
    messenger = PublishToOpenEphys()
    messenger.sendMessage(message)
    messenger.close()


class SubscribeToOpenEphys(object):
    """subscription-based zmq network event receiver

        This can be used to receive events published using the EventPublisher
        plugin.
    """

    def __init__(self, address='localhost', port=5557, timeout=2,
                 message_filter=''.encode(), verbose=True, save_messages=False):

        self.address = address
        self.port = port
        self.timeout = timeout
        self.message_filter = message_filter
        self.verbose = verbose
        self.save_messages = save_messages

        socket = None
        context = None
        context = zmq.Context()
        socket = context.socket(zmq.SUB)
        socket.RCVTIMEO = int(timeout * 1000)  # in milliseconds

        self.socket = socket
        self.context = context

        self.current_url = None
        self.thread = None
        self.messages = []
        self.lock = Lock()
        self.is_running = False
        self.callbacks = []

    def connect(self):

        if self.socket is None:
            return

        if self.is_connected():
            self.disconnect()

        url = "tcp://%s:%d" % (self.address, self.port)
        if self.verbose:
            print("Connecting subscriber to:", url)

        self.socket.connect(url)
        self.socket.setsockopt(zmq.SUBSCRIBE, self.message_filter)
        self.current_url = url

        self.is_running = True
        self.thread = Thread(target=self._run)
        self.thread.start()

    def disconnect(self):

        if self.socket is None:
            return

        if self.thread is not None and self.thread.is_alive():

            self.lock.acquire()
            self.is_running = False
            self.lock.release()

            self.thread.join()

        if self.is_connected():
            if self.verbose:
                print("Disconnecting subscriber from:", self.current_url)
            self.socket.disconnect(self.current_url)
            self.current_url = None

    def __del__(self):

        if self.socket is None:
            return

        if self.is_connected():
            if self.verbose:
                print("Disconnecting network subscriber ...")
            self.disconnect()

        if self.verbose:
            print("Terminating network context ...")
        self.socket.close()
        self.context.term()

    def is_connected(self):

        if self.socket is None:
            return False

        return self.current_url is not None

    def get_messages(self, clear=False):

        self.lock.acquire()
        msg = [m for m in self.messages]
        if clear:
            del self.messages[:]
        self.lock.release()

        return msg

    def add_callback(self, cb):

        self.callbacks.append(cb)

    def remove_callback(self, cb):

        if cb in self.callbacks:
            self.callbacks.remove(cb)

    def _send_message_to_callbacks(self, msg):

        for cb in self.callbacks:
            Thread(target=cb, args=(msg,)).start()

    def _run(self):

        while True:

            self.lock.acquire()
            running = self.is_running
            self.lock.release()

            if not running:
                break

            try:
                msg = self.socket.recv()

                if self.save_messages:
                    self.lock.acquire()
                    self.messages.append(msg)
                    self.lock.release()

                self._send_message_to_callbacks(msg.decode())

            except zmq.ZMQError:
                pass

            sleep(.01)

def SubscribeToOpenEphys_message_callback(msg):
    print("received event:", msg)

def SubscribeToOpenEphys_run_example(args):
    """
    An example script for using SubscribeToOpenEphys
    """

    # parse arguments
    address = 'localhost'  # address of system on which OE is running
    port = 5557  # port
    T = 10  # how long to listen for messages

    if len(args) > 1:
        address = args[1]
    if len(args) > 2:
        port = int(args[2])
    if len(args) > 3:
        T = float(args[3])

    try:
        # connect subscriber to event publisher
        sub = SubscribeToOpenEphys(address=address, port=port)
        sub.add_callback(SubscribeToOpenEphys_message_callback)
        sub.connect()

        # _run for T seconds
        sleep(T)

    except BaseException:
        traceback.print_exc()

    finally:
        # make sure the background thread is being stopped before exiting
        sub.disconnect()




from openEPhys_DACQ import NWBio
import matplotlib.pyplot as plt
import numpy as np
import argparse

def main(fpath):
    print('Opening NWB file...')
    data = NWBio.load_continuous(fpath)
    # Load timestamps into memory
    timestamps = np.array(data['timestamps'])
    timestamps = timestamps - timestamps[0]
    # Get user specified start and end time
    print('Session length is ' + str(timestamps[-1]) + ' seconds.')
    start_time = float(raw_input('Enter segment start time: '))
    end_time = float(raw_input('Enter segment end time: '))
    start_idx = np.argmin(np.abs(timestamps - start_time))
    end_idx = np.argmin(np.abs(timestamps - end_time))
    # Get user specified start and end time
    print('There are ' + str(data['continuous'].shape[1]) + ' channels.')
    first_chan = int(raw_input('Enter first channel nr: ')) - 1
    last_chan = int(raw_input('Enter last channel nr: '))
    # Load continuous data of specified shape
    print('Loading continuous data for segment...')
    timestamps = timestamps[start_idx:end_idx]
    continuous = np.array(data['continuous'][start_idx:end_idx, first_chan:last_chan])
    continuous = continuous.astype(np.float32) * 0.195
    n_channels = continuous.shape[1]
    n_datapoints = continuous.shape[0]
    data['file_handle'].close()
    # Convert continuous data such that it can all be show in single plot
    print('Converting continuous data for display...')
    mean_std = np.mean(np.std(continuous, axis=0))
    channel_spacing = mean_std * 3
    channel_spacer = np.linspace(0, (n_channels - 1) * channel_spacing, n_channels)
    channel_spacer = np.repeat(np.transpose(channel_spacer[:, None]), n_datapoints, axis=0)
    continuous = continuous + channel_spacer
    # Plot data
    print('Plotting...')
    fig = plt.figure()
    ax = plt.gca()
    plt.axis('off')
    plt.plot(timestamps, continuous, linewidth=1)
    plt.xlim(np.min(timestamps), np.max(timestamps))
    plt.ylim(np.min(continuous), np.max(continuous))
    plt.subplots_adjust(left=0, right=1, top=1, bottom=0)
    plt.gca().invert_yaxis()
    plt.show()

if __name__ == '__main__':
    # Input argument handling and help info
    parser = argparse.ArgumentParser(description='Display specific section of RAW data from NWB file.')
    parser.add_argument('fpath', type=str, nargs=1, 
                        help='Path to NWB file.')
    args = parser.parse_args()
    # Get paths to recording files
    fpath = args.fpath[0]
    main(fpath)



import pigpio
from ZMQcomms import paired_messenger
import argparse
from time import sleep

class GlobalClock_TTL_emitter(object):
    '''
    Produces very reliable pulses at regular intervals.
    Can be controlled over network with ZMQ.
    '''
    def __init__(self, ZMQcontrol=False, ZMQport=6000, ttlPin=18, frequency=10, pulse_width=1):
        '''
        ZMQcontrol - bool - default is False. Returns 'init_successful' over ZMQ and waits for commands.
        ZMQport - int - port to use for ZMQ communication at localhost.
        ttlPin - int - NOTE! Only hardware_PWM enabled pin can be used. See pigpio documentation.
        frequency - float - signal frequency (Hz)
        pulse_width - float - pulse width (milliseconds)
        '''
        self.ZMQcontrol = ZMQcontrol
        if self.ZMQcontrol:
            self.initialize_ZMQcomms(ZMQport)
        self.init_TTL_signalling(frequency=frequency, pulse_width=pulse_width, ttlPin=ttlPin)
        if self.ZMQcontrol:
            self.ZMQmessenger.sendMessage('init_successful')
            self.wait_for_ZMQ_commands()

    def initialize_ZMQcomms(self, port):
        '''
        Initializes ZMQ communication with Recording PC.
        '''
        self.ZMQmessenger = paired_messenger(port=int(port))
        self.ZMQmessenger.add_callback(self.command_parser)
        sleep(1) # This ensures all ZMQ protocols have been properly initated before finishing this process

    def command_parser(self, message):
        '''
        Parses incoming ZMQ message for function name, input arguments and calls that function.
        '''
        method_name = message.split(' ')[0]
        args = message.split(' ')[1:]
        getattr(self, method_name)(*args)

    @staticmethod
    def pwm_frequency_and_duty_cycle(signal_frequency, pulse_width):
        '''
        Returns PWM frequency and duty cycle to produce a signal that repeats at exactly signal_frequency

        frequency - float - pulse frequency (Hz)
        pulseWidth - float - pulse width (milliseconds)
        '''
        interval_between_pulse_starts = 1000.0 / float(signal_frequency)
        interval_between_pulse_end_and_start = interval_between_pulse_starts - float(pulse_width)
        pwm_frequency = int(round(1000.0 / interval_between_pulse_end_and_start))
        pwm_duty_cycle_ratio = float(pulse_width) / interval_between_pulse_end_and_start
        pwm_duty_cycle = int(1000000 * pwm_duty_cycle_ratio)

        return pwm_frequency, pwm_duty_cycle

    def init_TTL_signalling(self, frequency=10, pulse_width=1, ttlPin=18):
        '''
        Initializes RPi GPIO pin for sending regular TTL pulses.
        
        frequency - float - signal frequency (Hz)
        pulse_width - float - pulse width (milliseconds)
        ttlPin - int - NOTE! Only hardware_PWM enabled pin can be used. See pigpio documentation.
        '''
        self.ttlPin = ttlPin
        self.pwm_frequency, self.pwm_duty_cycle = self.pwm_frequency_and_duty_cycle(frequency, pulse_width)
        self.piPWM = pigpio.pi()
        self.piPWM.hardware_PWM(self.ttlPin, 0, 0)

    def wait_for_ZMQ_commands(self):
        '''
        Keeps the process alive so it can respond to ZMQ commands.
        '''
        self.keep_waiting_for_ZMQ_commands = True
        while self.keep_waiting_for_ZMQ_commands:
            sleep(0.1)

    def start(self):
        self.piPWM.hardware_PWM(self.ttlPin, self.pwm_frequency, self.pwm_duty_cycle)

    def stop(self):
        self.piPWM.hardware_PWM(self.ttlPin, 0, 0)

    def close(self):
        self.stop()
        self.piPWM.stop()
        if self.ZMQcontrol:
            self.ZMQmessenger.close()
            self.keep_waiting_for_ZMQ_commands = False


if __name__ == '__main__':
    # Input argument handling and help info.
    parser = argparse.ArgumentParser(description='Running this script initates GlobalClock_TTL_emitter class\n' + 
                                     'only if --remote flag is provided. Class waits for ZMQ commands.')
    parser.add_argument('--remote', action='store_true', 
                        help='Expects start, stop and close commands over ZMQ.')
    parser.add_argument('--port', type=int, nargs=1, 
                        help='The port to use for ZMQ paired_messenger with Recording PC.')
    args = parser.parse_args()
    # Only run the GlobalClock_TTL_emitter as a script if --remote flag provided.
    if args.remote:
        if args.port:
            Controller = GlobalClock_TTL_emitter(ZMQcontrol=True, ZMQport=args.port[0])
        else:
            print('--port was not provided, but is required to start the Class.')
    else:
        print('--remote flag was not used. Class can not be used as script without ZMQ commands.')




import os
from shutil import rmtree
import json


# TODO: KiloSort path currently hardcoded in StandardConfig.m. Should be here instead.


package_name = 'openEPhys_DACQ'

package_path = os.path.dirname(os.path.realpath(__file__))


class PackageConfiguration(object):

    config_file_path = os.path.join(os.path.expanduser('~'), '.{}_Config.json'.format(package_name))

    default_config = {
        'root_folder': os.path.join(os.path.expanduser('~'), 'RecordingData'),
        'recording_manager_settings_subfolder': 'RecordingManagerSettings',
        'klustakwik_path': 
            os.path.join(os.path.expanduser('~'), 'Programs', 'klustakwik', 'KlustaKwik')
    }

    def __init__(self):

        if os.path.isfile(self.config_file_path):

            with open(self.config_file_path, 'r') as config_file:
                self.config = json.loads(config_file.read())

        else:

            print('Configuration file for {} not found at expected location {}'.format(package_name,
                                                                                       self.config_file_path))
            print('Assuming this is the first time {} is run on this system'.format(package_name))
            print('Entering configuration mode...\n')

            self.configure_package()

            with open(self.config_file_path, 'w') as config_file:
                json.dump(self.config, config_file, sort_keys=True, indent=4)

            input('Configuration completed. Press Enter to continue:')

    def configure_root_folder(self):

        print('This program requires a working directory [root_folder] for:\n'
              + ' - Storing recorded data\n'
              + ' - Storing past settings of Recording Manager\n\n'
              + 'All existing data in the chosen working directory [root_folder]\n'
              + 'path will be deleted.\n'
              + 'The default path for working directory [root_folder] is {}'.format(self.config['root_folder']))

        decision = input('Use the default path for working directory [root_folder]? [y/N] ')

        if decision.lower() != 'y' and decision.lower() != 'yes':

            alternative_path = input('Please enter full path to alternative working directory [root_folder]: ')

            if os.path.isfile(alternative_path):

                raise ValueError('The entered working directory path {} leads to a file'.format(
                    os.path.dirname(alternative_path)
                ))

            elif os.path.isdir(os.path.dirname(alternative_path)):

                self.config['root_folder'] = alternative_path

            else:

                raise ValueError('The entered working directory path {} does not exist'.format(
                    os.path.dirname(alternative_path)
                ))

        if os.path.isdir(self.config['root_folder']):
            rmtree(self.config['root_folder'])
        os.mkdir(self.config['root_folder'])

        self.create_root_folder_subfolders()

        print('Working directory [root_folder] set as {}'.format(self.config['root_folder']))
        print('Recorded data will be found in directory {}'.format(os.path.join(self.config['root_folder'],
                                                                                'RecordingData')))

    def configure_klustakwik_path(self):
        
        print('To use KlustaKwik for processing tetrode recordings,'
              + 'working KlustaKwik installation is required')
        
        decision = input('Would you like to use the default KlustaKwik executable path\n'
                         + self.config['klustakwik_path']
                         + '[Y/n] ')

        if decision.lower() != 'y' and decision.lower() != 'yes':

            klustakwik_path = input('Please enter full path to KlustaKwik executable: ')

        else:

            klustakwik_path = self.config['klustakwik_path']

        if os.path.isfile(klustakwik_path):

            self.config['klustakwik_path'] = klustakwik_path

        else:

            raise ValueError('The entered KlustaKwik executable path {} does not lead to a file'.format(
                os.path.dirname(klustakwik_path)
            ))

    def create_root_folder_subfolders(self):

        os.mkdir(os.path.join(self.config['root_folder'],
                              self.config['recording_manager_settings_subfolder']))

    def configure_package(self):

        self.config = self.default_config

        self.configure_root_folder()

        self.configure_klustakwik_path()


def package_config():
    return PackageConfiguration().config



### This script is used to send commands to RPi over an SSH connection.

# Initiate connection by initating the class as connection = ssh(serverIP,sshUsername,sshPassword)
# Then send commands with connection.sendCommand('type_command_here')

from paramiko import client
from paramiko.ssh_exception import SSHException
from threading import Thread

class ssh:
    # This class was taken from:
    # https://daanlenaerts.com/blog/2016/01/02/python-and-ssh-sending-commands-over-ssh-using-paramiko/
    # 09/12/16
    client = None

    def __init__(self, address, username, password):
        self.address = address
        self.client = client.SSHClient()
        self.client.set_missing_host_key_policy(client.AutoAddPolicy())
        self.client.connect(address, username=username, password=password, look_for_keys=False)
        self.Ts_sendCommand = []

    def sendCommand(self, command, timeout=5, verbose=True):
        if(self.client):
            try:
                stdin, stdout, stderr = self.client.exec_command(command, timeout=timeout)
                if verbose:
                    while not stdout.channel.exit_status_ready():
                        # Print data when available
                        if stdout.channel.recv_ready():
                            alldata = stdout.channel.recv(1024)
                            prevdata = b"1"
                            while prevdata:
                                prevdata = stdout.channel.recv(1024)
                                alldata += prevdata
                            print(str(alldata))
            except SSHException as e:
                if e.__str__() == 'Timeout opening channel.':
                    print('ERROR: Command timeout!' + '\nFailed to send command: ' + command + ' | To: ' + self.address)
                elif e.__str__() == 'SSH session not active':
                    print('ERROR: Connection lost!' + '\nFailed to send command: ' + command + ' | To: ' + self.address)
                else:
                    raise e
        else:
            print("Connection not opened.")

    def sendCommand_threading(self, command, timeout=5, verbose=True):
        T = Thread(target=self.sendCommand, args=(command, timeout, verbose))
        T.start()
        self.Ts_sendCommand.append(T)

    def testConnection(self, timeout=5):
        '''
        Returns True if connection is active. False if connection inactive until timeout or session inactive.
        '''
        if(self.client):
            try:
                stdin, stdout, stderr = self.client.exec_command('ls', timeout=timeout)
                return True
            except SSHException as e:
                if e.__str__() == 'Timeout opening channel.':
                    return False
                elif e.__str__() == 'SSH session not active':
                    return False
                else:
                    raise e
        else:
            return False

    def disconnect(self, force=False):
        if(self.client):
            if not force and len(self.Ts_sendCommand) > 0:
                for T in self.Ts_sendCommand:
                    T.join()
            self.client.close()
        else:
            print("Connection not opened.")



# Python package for neural data acquisition with automated experiments
# Copyright (c) 2017-2020 Barry-lab https://www.ucl.ac.uk/biosciences/people/caswell-barry
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""A package for neural data acquisition with automated experiments"""

import os
import sys
from setuptools import setup, find_packages

if sys.version_info[0] != 3:
    raise ValueError('This package requires Python 3')

HERE = os.path.abspath(os.path.dirname(__file__))

# Workaround <http://www.eby-sarna.com/pipermail/peak/2010-May/003357.html>
try:
    import multiprocessing
except ImportError:
    pass

__project__      = 'openEPhys_DACQ'
__version__      = '0.79'
__author__       = 'Barry-lab'
__author_email__ = 'caswellbarry@ucl.ac.uk'
__url__          = 'https://barry-lab.github.io/openEPhys_DACQ/'

__classifiers__ = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: BSD License",
    "Operating System :: POSIX :: Linux",
    "Programming Language :: Python :: 3",
    "Topic :: Scientific/Engineering",
]

__keywords__ = [
    'open-ephys',
    'raspberrypi',
]

__requires__ = [
    'h5py',
    'matplotlib',
    'numpy',
    'opencv-python',
    'paramiko',
    'psutil',
    'pygame',
    'PyQt5',
    'pyqtgraph',
    'pyzmq',
    'scikit-learn',
    'scipy',
    'tqdm',
    'Pillow'
]

__extra_requires__ = {
    'doc':   ['sphinx' , 'sphinx_rtd_theme'],
    'Processing':  ['klustakwik', 'kilosort', 'matlab.engine']
}

__entry_points__ = {
    'console_scripts': [
        'openEPhys_RecordingManager = openEPhys_DACQ.RecordingManager:main',
        'openEPhys_Processing = openEPhys_DACQ.Processing:main'
    ]
}


def main():
    import io
    with io.open(os.path.join(HERE, 'README.md'), 'r') as readme:
        setup(
            name                 = __project__,
            version              = __version__,
            description          = __doc__,
            long_description     = readme.read(),
            long_description_content_type = 'text/markdown',
            classifiers          = __classifiers__,
            author               = __author__,
            author_email         = __author_email__,
            url                  = __url__,
            license              = [
                c.rsplit('::', 1)[1].strip()
                for c in __classifiers__
                if c.startswith('License ::')
            ][0],
            keywords             = __keywords__,
            packages             = find_packages(),
            include_package_data = True,
            install_requires     = __requires__,
            extras_require       = __extra_requires__,
            entry_points         = __entry_points__,
            python_requires      = '>=3',
        )


if __name__ == '__main__':
    main()



'''
This script creates Waveform GUI compatible files for all selected .waveform files.

Note, absolute values of position data and therefore speed information will be incorrect, 
due to issue with conversion of position data. Spatial correlograms work fine though.
'''

import os
import numpy as np
from scipy import interpolate
import subprocess
import openEPhys_DACQ.NWBio as NWBio
import openEPhys_DACQ.HelperFunctions as hfunct
from datetime import datetime
import argparse

from openEPhys_DACQ.package_configuration import package_config


def AxonaDataEEG_SamplingRate():
    return 250

def AxonaDataEGF_SamplingRate():
    return 4800

def interpolate_waveforms(waves, input_sampling_frequency=30000, 
                          output_sampling_frequency=48000, output_timestemps=50):
    '''
    Resamples waves to output_sampling_frequency.
    Input waves must have enough timesteps to cover output_timesteps.
    '''
    input_sample_step = 1000.0 / float(input_sampling_frequency)
    original_bins = np.arange(0.0, input_sample_step * waves.shape[1], 
                              input_sample_step).astype(np.float32)
    output_sample_step = 1000.0 / float(output_sampling_frequency)
    target_bins = np.arange(0.0, output_sample_step * output_timestemps, 
                            output_sample_step).astype(np.float32)
    if target_bins[-1] > original_bins[-1]:
        raise ValueError('Input waves do not have enough samples to interpolate requested output.')
    new_waves = np.zeros((waves.shape[0], output_timestemps), dtype=np.int8)
    for nwave in range(waves.shape[0]):
        # Interpolate each waveform
        interfunct = interpolate.interp1d(original_bins, waves[nwave,:].astype(np.float32))
        new_waves[nwave,:] = np.int8(np.round(interfunct(target_bins)))

    return new_waves

def create_DACQ_waveform_data_for_single_tetrode(spike_data_tet, data_time_edges, 
                                                 input_sampling_frequency=30000, 
                                                 output_sampling_frequency=48000, 
                                                 output_timestemps=50):
    '''
    spike_data - dict - with fields:
                 'waveforms' - numpy array of shape (n_spikes, 4, n_datapoints)
                               waveforms for all spikes on the 4 channels of the tetrode.
                 'timestamps' - numpy one dimensional array with spike times in seconds
    data_time_edges - tuple with two elements: start and end time of data (in seconds).
                      This is used with timestamps to discard spikes outside data range.

    Default values of the following are compatible with AxonaData:
        input_sampling_frequency - int or float - sampling frequency of input waveforms
        output_sampling_frequency - int or float - target sampling frequency of waveforms
        output_timestemps - int or float - number of timesteps required in output data

    Note: The length of input waveforms must be at least as long as output waveforms.
          In case of default values for output sampling frequency and timesteps, 
          the input waveforms need to be at least 1.03 milliseconds long.
    '''
    # Create DACQ data tetrode format
    waveform_data_dacq = []
    dacq_waveform_dtype = [('ts', '>i'), ('waveform', '50b')]
    dacq_waveform_waves_dtype = '>i1'
    dacq_waveform_timestamps_dtype = '>i'
    dacq_sampling_rate = 96000
    # Align timestamps to start from 0 at data_time_edges[0]
    spike_data_tet['timestamps'] = spike_data_tet['timestamps'] - data_time_edges[0]
    # Get waveforms
    waves = np.array(spike_data_tet['waveforms'], dtype=np.float32)
    waves = -waves
    waves = np.swapaxes(waves,1,2)
    nspikes = waves.shape[0]
    # Set waveforms values on this tetrode to range -127 to 127
    if nspikes > 1:
        mean_waves = np.mean(waves, axis=0)
        wave_peak = mean_waves.max()
        maxchan = np.where(mean_waves == wave_peak)[1][0]
        maxidx = np.where(mean_waves == wave_peak)[0][0]
        wave_peak_std = np.std(waves[:,maxidx,maxchan])
        max_range = wave_peak + 4 * wave_peak_std # This sets int8 range
        waves = waves / max_range
        waves = waves * 127
        waves[waves > 127] = 127
        waves[waves < -127] = -127
    waves = waves.astype(np.int8)
    # Where channels are missing, add 0 values to waveform values
    if waves.shape[2] < 4:
        waves = np.concatenate((waves, np.zeros((waves.shape[0], waves.shape[1], 4 - waves.shape[2]))), axis=2)
    # Reshape 3D waveform matrix into 2D matrix such that waveforms for 
    # first spike from all four channels are on consecutive rows.
    waves = np.reshape(np.ravel(np.transpose(waves, (0, 2, 1)), 'C'), (nspikes * 4, waves.shape[1]))
    # Interpolate waveforms to 48000 Hz resolution
    waves = interpolate_waveforms(waves, input_sampling_frequency, 
                                  output_sampling_frequency, output_timestemps)
    # Create DACQ datatype structured array
    waveform_data_dacq = np.zeros(nspikes * 4, dtype=dacq_waveform_dtype)
    # Input waveform values, leaving a trailing end of zeros due to lower sampling rate
    waves = waves.astype(dtype=dacq_waveform_waves_dtype)
    waveform_data_dacq['waveform'][:,:waves.shape[1]] = waves
    # Arrange timestamps into a vector where timestamp for a spike is
    # repeated for each of the 4 channels
    timestamps = np.array(spike_data_tet['timestamps'])
    timestamps = np.ravel(np.repeat(np.reshape(timestamps, (len(timestamps), 1)), 4, axis=1),'C')
    # Convert OpenEphys timestamp sampling rate to DACQ sampling rate
    timestamps = timestamps * float(dacq_sampling_rate)
    timestamps_dacq = np.round(timestamps).astype(dtype=dacq_waveform_timestamps_dtype)
    # Input timestamp values to the dacq data matrix
    waveform_data_dacq['ts'] = timestamps_dacq
    
    return waveform_data_dacq

def create_DACQ_waveform_data(spike_data, data_time_edges, 
                              input_sampling_frequency=30000, 
                              output_sampling_frequency=48000, 
                              output_timestemps=50):
    '''
    This is a function to create_AxonaData waveforms using multiprocessing.
    See create_DACQ_waveform_data_for_single_tetrode() for description of input arguments.

    spike_data - list of spike_data_tet - see create_DACQ_waveform_data_for_single_tetrode()
    '''
    input_args = []
    for spike_data_tet in spike_data:
        input_args.append((spike_data_tet, data_time_edges, input_sampling_frequency, 
                           output_sampling_frequency, output_timestemps))
    multiprocessor = hfunct.multiprocess()
    waveform_data_dacq = multiprocessor.map(create_DACQ_waveform_data_for_single_tetrode, 
                                            len(input_args), args_list=input_args, 
                                            max_memory_usage=0.66)

    return waveform_data_dacq

def AxonaData_max_pix_value():
    return 600

def convert_position_data_from_cm_to_pixels(xy_pos):
    '''
    Converts cm positions to pixel values in AxonaData acceptable range
    Computes the conversion pixels_per_metre value
    '''
    max_pix = AxonaData_max_pix_value()
    xy_pos = xy_pos - np.nanmin(xy_pos)
    max_cm = np.nanmax(xy_pos)
    pix_per_cm = float(max_pix) / float(max_cm)
    xy_pos = xy_pos * pix_per_cm
    #pixels_per_metre = int(np.round(pix_per_cm * 100))
    pixels_per_metre = max_pix

    return xy_pos, pixels_per_metre

def create_DACQ_pos_data(posdata, data_time_edges=None, pixels_per_metre=None, dacq_pos_samplingRate=50):
    '''
    posdata - numpy array as type np.float32 with shape (N x 5) - 
              where there are N samples and colums are:
              timestamp (s), xpos1, ypos1, xpos2, ypos2
              Missing datapoints should be NaN. If no second LED, then NaN columns.

    data_time_edges - tuple with two elements: start and end time of synthetic position data (in seconds)
    
    If pixels_per_metre is entered, it is assumed that ProcessedPos data is
    in pixel values, not centimeters. Otherwise, position data values are scaled 
    into range [0, AxonaData_max_pix_value()] and pixels_per_metre set to AxonaData_max_pix_value().

    dacq_pos_samplingRate - float or int - sampling rate of AxonaData position data
                            default value is compatible with AxonaData.
    '''
    # Create DACQ data pos format
    dacq_pos_dtype = [('ts', '>i'), ('pos', '>8h')]
    # dacq_pos_timestamp_dtype = '>i4'
    dacq_pos_xypos_dtype = '>i2'
    dacq_pos_timestamp_dtype = '>i'
    if posdata is None:
        print('Warning! No position data provided, creating fake position data for AxonaData.')
        if data_time_edges is None:
            raise ValueError('data_time_edges as start and end of recording in seconds \n' + \
                             'is required to create synthetic data')
        countstamps = np.arange(np.floor((data_time_edges[1] - data_time_edges[0]) * float(dacq_pos_samplingRate)))
        dacq_timestamps = np.float64(countstamps) / float(dacq_pos_samplingRate)
        xy_pos = np.repeat(np.linspace(0, 100, dacq_timestamps.size)[:, None], 4, axis=1).astype(np.float32)
    else:
        # Extract xy values and timestamps from posdata
        xy_pos = posdata[:,1:5].astype(np.float32)
        timestamps = posdata[:,0].astype(np.float32)
        # Crop position data outside data_time_edges
        idx_outside_data_time = timestamps < data_time_edges[0]
        idx_outside_data_time = np.logical_or(idx_outside_data_time, timestamps > data_time_edges[1])
        idx_outside_data_time = np.where(idx_outside_data_time)[0]
        timestamps = np.delete(timestamps, idx_outside_data_time, 0)
        xy_pos = np.delete(xy_pos, idx_outside_data_time, 0)
        # Realign position data start to 0 at data_time_edges[0]
        timestamps = timestamps - data_time_edges[0]
        # Set minumum value of x and y to be 0
        x_min = np.nanmin(xy_pos[:, 0])
        y_min = np.nanmin(xy_pos[:, 1])
        if not all(np.isnan(xy_pos[:, 2])):
            x_min = min(x_min, np.nanmin(xy_pos[:, 2]))
        if not all(np.isnan(xy_pos[:, 3])):
            y_min = min(y_min, np.nanmin(xy_pos[:, 3]))
        xy_pos[:, 0] = xy_pos[:, 0] - x_min
        xy_pos[:, 2] = xy_pos[:, 2] - x_min
        xy_pos[:, 1] = xy_pos[:, 1] - y_min
        xy_pos[:, 3] = xy_pos[:, 3] - y_min
        # Interpolate position data to 50Hz
        countstamps = np.arange(np.floor(timestamps[-1] * float(dacq_pos_samplingRate)))
        dacq_timestamps = np.float64(countstamps) / float(dacq_pos_samplingRate)
        xy_pos_interp = np.zeros((dacq_timestamps.size, xy_pos.shape[1]), dtype=np.float32)
        for ncoord in range(xy_pos.shape[1]):
            xy_pos_interp[:,ncoord] = np.interp(dacq_timestamps, timestamps, xy_pos[:,ncoord])
        xy_pos = xy_pos_interp
    # If no pixels_per_metre provided, convert position data to pixel values
    # and calculate pixels_per_metre for range 600
    if pixels_per_metre is None:
        xy_pos, pixels_per_metre = convert_position_data_from_cm_to_pixels(xy_pos)
    # Round values and turn into integers
    xy_pos = np.int16(np.round(xy_pos))
    # Find minimum and maximum values for x and y values
    limits = {'x_max': max([np.nanmax(xy_pos[:, 0]), np.nanmax(xy_pos[:, 2])]), 
              'x_min': min([np.nanmin(xy_pos[:, 0]), np.nanmin(xy_pos[:, 2])]), 
              'y_max': max([np.nanmax(xy_pos[:, 1]), np.nanmax(xy_pos[:, 3])]), 
              'y_min': min([np.nanmin(xy_pos[:, 1]), np.nanmin(xy_pos[:, 3])])}
    # Set NaN values to 1023
    xy_pos[np.isnan(xy_pos)] = 1023
    # Convert datatypes to DACQ data type
    dacq_timestamps = dacq_timestamps.astype(dtype=dacq_pos_timestamp_dtype)
    xy_pos = xy_pos.astype(dtype=dacq_pos_xypos_dtype)
    # Arrange xy_pos as in the DACQ data matrix including the dot size values
    xy_pos = np.concatenate((xy_pos, 20 * np.ones((xy_pos.shape[0], 1), dtype=dacq_pos_xypos_dtype)), axis=1)
    xy_pos = np.concatenate((xy_pos, 10 * np.ones((xy_pos.shape[0], 1), dtype=dacq_pos_xypos_dtype)), axis=1)
    xy_pos = np.concatenate((xy_pos, 30 * np.ones((xy_pos.shape[0], 1), dtype=dacq_pos_xypos_dtype)), axis=1)
    xy_pos = np.concatenate((xy_pos, np.zeros((xy_pos.shape[0], 1), dtype=dacq_pos_xypos_dtype)), axis=1)
    # Create DACQ datatype structured array
    pos_data_dacq = np.zeros(dacq_timestamps.size, dtype=dacq_pos_dtype)
    # Input timestamps and pos_xy to DACQ array
    pos_data_dacq['ts'] = dacq_timestamps
    pos_data_dacq['pos'] = xy_pos

    return pos_data_dacq, pixels_per_metre, limits

def create_DACQ_eeg_or_egf_data(eeg_or_egf, data, data_time_edges, target_range=1000):
    '''
    EEG is lowpass filtered to half the target sampling rate, 
    downsampled to dacq_eeg_samplingrate and 
    inverted to same polarity as spikes in AxonaFormat.
    The data is also clipped to specified range in values and time.

    The returned array is in correct format to written to binary Axona file.

    eeg_or_egf - str - 'eeg' or 'egf' specifies ouput data format and sampling rate
    data - dict with following fields:
           'data' - numpy  array in dtype=numpy.float32 with dimensions (N x n_chan) - 
                  LFP data to convert.
           'timestamps' - numpy one dimensional array in dtype=numpy.float32 - 
                        timestamps in seconds for each of the datapoints in data
           'sampling_rate' - int or float - sampling rate of data
    data_time_edges - tuple with two elements: start and end time of data (in seconds).
                      This is used with timestamps to crop EEG data outside data range.
    target_range - int or float - EEG data with voltage values above this will be clipped.
    '''
    # AxonaData eeg data parameters
    if eeg_or_egf == 'eeg':
        output_SamplingRate = AxonaDataEEG_SamplingRate()
    elif eeg_or_egf == 'egf':
        output_SamplingRate = AxonaDataEGF_SamplingRate()
    if output_SamplingRate > data['sampling_rate']:
        raise ValueError('Input data sampling rate is lower than requested output data.')
    if data['data'].dtype  != np.float32:
        raise ValueError('Input data dtype is not numpy.float32.')
    lowpass_frequency = output_SamplingRate / 2.0
    # Filter data with lowpass butter filter
    data_in_processing = []
    for n_chan in range(data['data'].shape[1]):
        data_in_processing.append(hfunct.butter_lowpass_filter(data['data'][:, n_chan].copy(), 
                                                               sampling_rate=float(data['sampling_rate']), 
                                                               lowpass_frequency=lowpass_frequency, 
                                                               filt_order=4))
    # Crop data outside data_time_edges
    idx_outside_data_time = data['timestamps'] < data_time_edges[0]
    idx_outside_data_time = np.logical_or(idx_outside_data_time, data['timestamps'] > data_time_edges[1])
    idx_outside_data_time = np.where(idx_outside_data_time)[0]
    cropped_timestamps = np.delete(data['timestamps'].copy(), idx_outside_data_time, 0)
    for n_chan in range(len(data_in_processing)):
        data_in_processing[n_chan] = np.delete(data_in_processing[n_chan], idx_outside_data_time, 0)
    # Resample data to dacq_eeg sampling rate
    original_timestamps = (cropped_timestamps - cropped_timestamps[0])
    target_timestamps = np.arange(0, original_timestamps[-1], 1.0 / float(output_SamplingRate))
    for n_chan in range(len(data_in_processing)):
        interfunct = interpolate.interp1d(original_timestamps, data_in_processing[n_chan])
        data_in_processing[n_chan] = interfunct(target_timestamps)
    # Invert data
    for n_chan in range(len(data_in_processing)):
        data_in_processing[n_chan] = -data_in_processing[n_chan]
    # Adjust EEG data format and range
    for n_chan in range(len(data_in_processing)):
        data_in_processing[n_chan] = data_in_processing[n_chan] - np.mean(data_in_processing[n_chan])
    for n_chan in range(len(data_in_processing)):
        data_in_processing[n_chan] = data_in_processing[n_chan] / target_range
    for n_chan in range(len(data_in_processing)):
        data_in_processing[n_chan] = data_in_processing[n_chan] * 127
    for n_chan in range(len(data_in_processing)):
        data_in_processing[n_chan][data_in_processing[n_chan] > 127] = 127
        data_in_processing[n_chan][data_in_processing[n_chan] < -127] = -127
    # Create DACQ data eeg format
    if eeg_or_egf == 'eeg':
        dacq_eeg_dtype = [('eeg', '=b')]
        dacq_eeg_data_dtype = '=b'
    elif eeg_or_egf == 'egf':
        dacq_eeg_dtype = [('eeg', np.int16)]
        dacq_eeg_data_dtype = np.int16
    eeg_data_dacq = [None] * len(data_in_processing)
    for n_chan in range(len(data_in_processing)):
        dacq_eeg = data_in_processing[n_chan].astype(dtype=dacq_eeg_data_dtype)
        eeg_data_dacq[n_chan] = np.zeros(dacq_eeg.size, dtype=dacq_eeg_dtype)
        eeg_data_dacq[n_chan]['eeg'] = dacq_eeg

    return eeg_data_dacq

def create_DACQ_eeg_data(data, data_time_edges, target_range=1000):
    return create_DACQ_eeg_or_egf_data('eeg', data, data_time_edges, 
                                       target_range=target_range)

def create_DACQ_egf_data(data, data_time_edges, target_range=1000):
    return create_DACQ_eeg_or_egf_data('egf', data, data_time_edges, 
                                       target_range=target_range)

def header_templates(htype):
    # This function returns the default header necessary for the waveform and pos files
    if htype == 'waveforms':
        keyorder = ['trial_date', 'trial_time', 'experimenter', 'comments', \
              'duration', 'sw_version', 'num_chans', 'timebase', \
              'bytes_per_timestamp', 'samples_per_spike', 'sample_rate', \
              'bytes_per_sample', 'spike_format', 'num_spikes']

        header = {'trial_date': 'Thursday, 3 Mar 2016', 
                  'trial_time': '11:28:24', 
                  'experimenter': 'ST', 
                  'comments': 'R2410', 
                  'duration': '1', 
                  'sw_version': '1.2.2.14', 
                  'num_chans': '4', 
                  'timebase': '96000 hz', 
                  'bytes_per_timestamp': '4', 
                  'samples_per_spike': '50', 
                  'sample_rate': '48000 hz',
                  'bytes_per_sample': '1', 
                  'spike_format': 't,ch1,t,ch2,t,ch3,t,ch4', 
                  'num_spikes': '0'}

    elif htype == 'pos':
        keyorder = ['trial_date', 'trial_time', 'experimenter', 'comments', \
              'duration', 'sw_version', 'num_colours', 'min_x', 'max_x', \
              'min_y', 'max_y', 'window_min_x', 'window_max_x', \
              'window_min_y', 'window_max_y', 'timebase', \
              'bytes_per_timestamp', 'sample_rate', \
              'EEG_samples_per_position', 'bearing_colour_1', \
              'bearing_colour_2', 'bearing_colour_3', 'bearing_colour_4', \
              'pos_format', 'bytes_per_coord', 'pixels_per_metre', \
              'num_pos_samples']

        header = {'trial_date': 'Thursday, 3 Mar 2016', 
                  'trial_time': '11:28:24', 
                  'experimenter': 'ST', 
                  'comments': 'R2410', 
                  'duration': '1', 
                  'sw_version': '1.2.2.14', 
                  'num_colours': '4', 
                  'min_x': '0', 
                  'max_x': str(AxonaData_max_pix_value()), 
                  'min_y': '0', 
                  'max_y': str(AxonaData_max_pix_value()), 
                  'window_min_x':'0', 
                  'window_max_x': str(AxonaData_max_pix_value()), 
                  'window_min_y': '0', 
                  'window_max_y': str(AxonaData_max_pix_value()), 
                  'timebase': '50 hz', 
                  'bytes_per_timestamp': '4', 
                  'sample_rate': '50.0 hz', 
                  'EEG_samples_per_position': '5', 
                  'bearing_colour_1': '0', 
                  'bearing_colour_2': '0', 
                  'bearing_colour_3': '0', 
                  'bearing_colour_4': '0', 
                  'pos_format': 't,x1,y1,x2,y2,numpix1,numpix2', 
                  'bytes_per_coord': '2', 
                  'pixels_per_metre': '100', 
                  'num_pos_samples': '0'}
    
    elif htype == 'eeg':
        keyorder = ['trial_date', 'trial_time', 'experimenter', 'comments', \
              'duration', 'sw_version', 'num_chans', 'sample_rate', 'EEG_samples_per_position', \
              'bytes_per_sample', 'num_EEG_samples']

        header = {'trial_date': 'Tuesday, 14 Dec 2010', 
                  'trial_time': '10:44:07', 
                  'experimenter': 'cb', 
                  'comments': '1838', 
                  'duration': '1201', 
                  'sw_version': '1.0.2', 
                  'num_chans': '1', 
                  'sample_rate': '250.0 hz', 
                  'EEG_samples_per_position': '5', 
                  'bytes_per_sample': '1', 
                  'num_EEG_samples': '0'}
    
    elif htype == 'egf':
        keyorder = ['trial_date', 'trial_time', 'experimenter', 'comments', \
              'duration', 'sw_version', 'num_chans', 'sample_rate', \
              'bytes_per_sample', 'num_EGF_samples']

        header = {'trial_date': 'Tuesday, 14 Dec 2010', 
                  'trial_time': '10:44:07', 
                  'experimenter': 'cb', 
                  'comments': '1838', 
                  'duration': '1201', 
                  'sw_version': '1.0.2', 
                  'num_chans': '1', 
                  'sample_rate': '4800 hz', 
                  'bytes_per_sample': '2', 
                  'num_EGF_samples': '0'}
                  
    return header, keyorder

def update_header(header, experiment_info, new_values_dict):
    header['trial_date'] = experiment_info['trial_date']
    header['trial_time'] = experiment_info['trial_time']
    header['comments'] = experiment_info['animal']
    header['duration'] = experiment_info['duration']
    for key in new_values_dict.keys():
        header[key] = new_values_dict[key]

    return header

def getExperimentInfo(fpath=None):
    '''
    Extract experiment info from NWB file and sets to correc tformat for
    createAxonaData() function.

    If fpath is not provided, sham data is provided that is compatible
    with createAxonaData() function.
    '''
    if not (fpath is None) and NWBio.check_if_settings_available(fpath,'/General/animal/'):
        animal = NWBio.load_settings(fpath,'/General/animal/')
    else:
        animal = 'unknown'
    if not (fpath is None) and NWBio.check_if_settings_available(fpath,'/Time/'):
        full_timestring = NWBio.load_settings(fpath,'/Time/')
        timeobject = datetime.strptime(full_timestring, '%Y-%m-%d_%H-%M-%S')
    else:
        timeobject = datetime.now()
    datestring = timeobject.strftime('%d-%m-%y')
    timestring = timeobject.strftime('%H-%M-%S')
    experiment_info = {'trial_date': datestring, 
                       'trial_time': timestring, 
                       'animal': animal}
    
    return experiment_info

def get_first_available_spike_data(OpenEphysDataPath, tetrode_nrs, use_idx_keep, use_badChan, 
                                   clustering_name=None):
    _, spike_names = NWBio.processing_method_and_spike_name_combinations()
    spike_data_available = False
    for spike_name in spike_names:
        spike_data = NWBio.load_spikes(OpenEphysDataPath, tetrode_nrs=tetrode_nrs, 
                                       spike_name=spike_name, use_idx_keep=use_idx_keep, 
                                       use_badChan=use_badChan, clustering_name=clustering_name)
        if len(spike_data) > 0:
            spike_data_available = True
            break
    if not spike_data_available:
        raise Exception('Spike data is not available in file: ' + OpenEphysDataPath + '\nChecked for following spike_name: ' + str(spike_names))

    return spike_data

def write_file_in_axona_format(filename, header, header_keyorder, data):
    '''
    Writes data in axona format
    '''
    # Set data start and end tokens
    DATA_START_TOKEN = 'data_start'
    DATA_END_TOKEN = '\r\ndata_end\r\n'
    with open(filename, 'wb') as f:
        # Write header in the correct order
        for key in header_keyorder:
            if 'num_spikes' in key:
                # Replicate spaces following num_spikes in original dacq files
                stringval = header[key]
                while len(stringval) < 10:
                    stringval += ' '
                f.write(hfunct.b(key + ' ' + stringval + '\r\n'))
            elif 'num_pos_samples' in key:
                # Replicate spaces following num_pos_samples in original dacq files
                stringval = header[key]
                while len(stringval) < 10:
                    stringval += ' '
                f.write(hfunct.b(key + ' ' + stringval + '\r\n'))
            elif 'duration' in key:
                # Replicate spaces following duration in original dacq files
                stringval = header[key]
                while len(stringval) < 10:
                    stringval += ' '
                f.write(hfunct.b(key + ' ' + stringval + '\r\n'))
            else:
                f.write(hfunct.b(key + ' ' + header[key] + '\r\n'))
        # Write the start token string
        f.write(hfunct.b(DATA_START_TOKEN))
        # Write the data into the file in binary format
        data.tofile(f)
        # Write the end token string
        f.write(hfunct.b(DATA_END_TOKEN))


def write_clusterIDs_in_CLU_format(clusterIDs, cluFileName):
    lines = [hfunct.b(str(max(clusterIDs)) + '\r\n')]
    for nclu in list(clusterIDs):
        lines.append(hfunct.b(str(nclu) + '\r\n'))
    with open(cluFileName, 'wb') as file:
        file.writelines(lines)

def write_set_file(setFileName, new_values_dict):
    sourcefile = os.path.join(package_config.package_path, 'Utils', 'SetFileBase.set')
    # Read in base .set file
    with open(sourcefile, 'rb') as file:
        lines = file.readlines()
    # Correct lines based on new_values_dict
    for key in new_values_dict.keys():
        for nl, line in enumerate(lines):
            if hfunct.b(key + ' ') in line:
                lines[nl] = hfunct.b(key + ' ' + new_values_dict[key] + '\r\n')
                break
    # Write the .set file with corrected lines
    with open(setFileName, 'wb') as file:
        file.writelines(lines)

def load_eegData(fpath, eegChans, bitVolts=0.195):
    # Load EEG data of selected channel
    print([hfunct.time_string(), 'DEBUG: load from file'])
    continuous_data = NWBio.load_continuous_as_array(fpath, eegChans)
    if not (continuous_data is None):
        print([hfunct.time_string(), 'DEBUG: extract timestamps'])
        timestamps = np.array(continuous_data['timestamps'])
        print([hfunct.time_string(), 'DEBUG: transforming into float32'])
        data = continuous_data['continuous'].astype(np.float32)
        sampling_rate = NWBio.OpenEphys_SamplingRate()
    else:
        # If no raw data available, load downsampled data for that tetrode
        lowpass_data = NWBio.load_tetrode_lowpass(fpath)
        timestamps = np.array(lowpass_data['tetrode_lowpass_timestamps'])
        sampling_rate = NWBio.OpenEphys_SamplingRate() / float(lowpass_downsampling)
        if not isinstance(eegChans, list):
            eegChans = [eegChans]
        tetrodes = []
        for eegChan in eegChans:
            tetrodes.append(hfunct.channels_tetrode(eegChan))
        data = np.array(lowpass_data['tetrode_lowpass'][:, tetrodes]).astype(np.float32)
    # Scale data with bitVolts value to convert from raw data to voltage values
    print([hfunct.time_string(), 'DEBUG: converting to bitVolts'])
    data = data * bitVolts

    return {'data': data, 'timestamps': timestamps, 'sampling_rate': sampling_rate}

def createAxonaData_for_NWBfile(OpenEphysDataPath, spike_name='first_available', channel_map=None, 
                                subfolder='AxonaData', eegChans=None, pixels_per_metre=None, 
                                show_output=False, clustering_name=None):
    # Construct path for AxonaData and get experiment info
    AxonaDataPath = os.path.join(os.path.dirname(OpenEphysDataPath), subfolder)
    experiment_info = getExperimentInfo(OpenEphysDataPath)
    # Get channel_map for this dataset
    if channel_map is None:
        if NWBio.check_if_settings_available(OpenEphysDataPath,'/General/channel_map/'):
            channel_map = NWBio.load_settings(OpenEphysDataPath,'/General/channel_map/')
        else:
            raise ValueError('Channel map could not be generated. Enter channels to process.')
    # Get position data for this recording
    data_time_edges = NWBio.get_processed_tracking_data_timestamp_edges(OpenEphysDataPath)
    if NWBio.check_if_processed_position_data_available(OpenEphysDataPath):
        posdata = NWBio.load_processed_tracking_data(OpenEphysDataPath)
    else:
        posdata = None
    # Create AxonaData separately for each recording area
    for area in channel_map.keys():
        # Load spike data
        tetrode_nrs = hfunct.get_tetrode_nrs(channel_map[area]['list'])
        print('Loading spikes for tetrodes nr: ' +  ', '.join(map(str, tetrode_nrs)))
        if spike_name == 'first_available':
            spike_data = get_first_available_spike_data(OpenEphysDataPath, tetrode_nrs, 
                                                        use_idx_keep=True, use_badChan=True,
                                                        clustering_name=clustering_name)
        else:
            spike_data = NWBio.load_spikes(OpenEphysDataPath, tetrode_nrs=tetrode_nrs, 
                                           spike_name=spike_name, use_idx_keep=True, 
                                           use_badChan=True, clustering_name=clustering_name)
        # Load eeg data
        if eegChans is None:
            eegData = None
        else:
            eegChansInArea_Bool = [x in channel_map[area]['list'] for x in eegChans]
            if any(eegChansInArea_Bool):
                eegChansInArea = [x for (x,y) in zip(eegChans, eegChansInArea_Bool) if y]
                print('Loading LFP data for channels: ' +  ', '.join(map(str, eegChansInArea)))
                eegData = load_eegData(OpenEphysDataPath, eegChansInArea)
            else:
                eegData = None
        createAxonaData(AxonaDataPath, spike_data, data_time_edges, posdata=posdata, 
                        experiment_info=experiment_info, axona_file_name=area, 
                        eegData=eegData, pixels_per_metre=pixels_per_metre, 
                        show_output=show_output)

def concatenate_posdata_across_recordings(posdata, data_time_edges, recording_edges):
    # Only keep timestamps and x,y for led1 and led2
    posdata = [x[:, :5] for x in posdata]
    for n_rec in range(len(posdata)):
        # Crop data outside data_time_edges and transform timestamps to continuous recording
        idx_outside_data_time = posdata[n_rec][:, 0] < data_time_edges[n_rec][0]
        idx_outside_data_time = np.logical_or(idx_outside_data_time, 
                                              posdata[n_rec][:, 0] > data_time_edges[n_rec][1])
        idx_outside_data_time = np.where(idx_outside_data_time)[0]
        posdata[n_rec] = np.delete(posdata[n_rec], idx_outside_data_time, axis=0)
        posdata[n_rec][:, 0] = posdata[n_rec][:, 0] - data_time_edges[n_rec][0] + recording_edges[n_rec][0]
    # Concatenate posdata
    posdata = np.concatenate(posdata, axis=0)

    return posdata

def concatenate_spike_data_across_recordings(spike_data, data_time_edges, recording_edges):
    new_spike_data = [None] * len(spike_data[0])
    for n_tet in range(len(new_spike_data)):
        print([hfunct.time_string(), 'DEBUG: concatenating data for tetrode ', n_tet])
        waveforms = [data[n_tet]['waveforms'] for data in spike_data]
        clusterIDs = [data[n_tet]['clusterIDs'] for data in spike_data]
        timestamps = [data[n_tet]['timestamps'] for data in spike_data]
        for n_rec in range(len(timestamps)):
            # Transform timestamps to continuous recordings
            timestamps[n_rec] = timestamps[n_rec] - data_time_edges[n_rec][0] + recording_edges[n_rec][0]
        print([hfunct.time_string(), 'DEBUG: The concatenation'])
        new_spike_data[n_tet] = {'waveforms': np.concatenate(waveforms, axis=0), 
                                 'clusterIDs': np.concatenate(clusterIDs, axis=0), 
                                 'timestamps': np.concatenate(timestamps, axis=0)}

    return new_spike_data

def concatenate_eegData_across_recordings(eegData, data_time_edges, recording_edges):
    for n_rec in range(len(eegData)):
        # Crop data outside data_time_edges and transform timestamps to continuous recording
        idx_outside_data_time = eegData[n_rec]['timestamps'] < data_time_edges[n_rec][0]
        idx_outside_data_time = np.logical_or(idx_outside_data_time, 
                                              eegData[n_rec]['timestamps'] > data_time_edges[n_rec][1])
        idx_outside_data_time = np.where(idx_outside_data_time)[0]
        eegData[n_rec]['timestamps'] = np.delete(eegData[n_rec]['timestamps'], idx_outside_data_time, axis=0)
        eegData[n_rec]['timestamps'] = eegData[n_rec]['timestamps'] - data_time_edges[n_rec][0] + recording_edges[n_rec][0]
        eegData[n_rec]['data'] = np.delete(eegData[n_rec]['data'], idx_outside_data_time, axis=0)
    # Concatenate data and timestamps
    print([hfunct.time_string(), 'DEBUG: The concatenation'])
    new_eegData = {'data': np.concatenate([x['data'] for x in eegData], axis=0), 
                   'timestamps': np.concatenate([x['timestamps'] for x in eegData], axis=0), 
                   'sampling_rate': eegData[0]['sampling_rate']}

    return new_eegData

def createAxonaData_for_multiple_NWBfiles(OpenEphysDataPaths, AxonaDataPath, 
                                          spike_name='first_available', channel_map=None, 
                                          eegChans=None, pixels_per_metre=None, 
                                          show_output=False, clustering_name=None):
    # Get experiment info
    if len(OpenEphysDataPaths) > 1:
        print('Using experiment_info from first recording only.')
    experiment_info = getExperimentInfo(OpenEphysDataPaths[0])
    # Get channel_map for this dataset
    if channel_map is None:
        if len(OpenEphysDataPaths) > 1:
            print('Using channel_map from first recording only.')
        if NWBio.check_if_settings_available(OpenEphysDataPaths[0],'/General/channel_map/'):
            channel_map = NWBio.load_settings(OpenEphysDataPaths[0],'/General/channel_map/')
        else:
            raise ValueError('Channel map could not be generated. Enter channels to process.')
    # Compute start and end times of each segment of the recording
    data_time_edges = []
    for OpenEphysDataPath in OpenEphysDataPaths:
        data_time_edges.append(NWBio.get_processed_tracking_data_timestamp_edges(OpenEphysDataPath))
    recording_edges = []
    recording_duration = 0
    for dte in data_time_edges:
        end_of_this_recording = recording_duration + (dte[1] - dte[0])
        recording_edges.append([recording_duration, end_of_this_recording])
        recording_duration = end_of_this_recording
    combined_data_time_edges = [recording_edges[0][0], recording_edges[-1][1]]
    # Get position data for these recordings
    print('Loading position data.')
    posdata = []
    for OpenEphysDataPath in OpenEphysDataPaths:
        if NWBio.check_if_processed_position_data_available(OpenEphysDataPath):
            posdata.append(NWBio.load_processed_tracking_data(OpenEphysDataPath))
        else:
            posdata.append(None)
    if any([x is None for x in posdata]):
        posdata = None
    else:
        posdata = concatenate_posdata_across_recordings(posdata, data_time_edges, 
                                                        recording_edges)
    # Create AxonaData separately for each recording area
    for area in channel_map.keys():
        # Load spike data
        tetrode_nrs = hfunct.get_tetrode_nrs(channel_map[area]['list'])
        print('Loading spikes for tetrodes nr: ' +  ', '.join(map(str, tetrode_nrs)))
        spike_data = []
        for OpenEphysDataPath in OpenEphysDataPaths:
            if spike_name == 'first_available':
                spike_data.append(get_first_available_spike_data(OpenEphysDataPath, tetrode_nrs, 
                                                                 use_idx_keep=True, use_badChan=True))
            else:
                print([hfunct.time_string(), 'DEBUG: loading spikes of tet ', tetrode_nrs, ' from ', OpenEphysDataPath])
                spike_data.append(NWBio.load_spikes(OpenEphysDataPath, tetrode_nrs=tetrode_nrs, 
                                                    spike_name=spike_name, use_idx_keep=True, 
                                                    use_badChan=True, clustering_name=clustering_name))
        spike_data = concatenate_spike_data_across_recordings(spike_data, data_time_edges, 
                                                              recording_edges)
        # Load eeg data
        if eegChans is None:
            eegData = None
        else:
            eegChansInArea_Bool = [x in channel_map[area]['list'] for x in eegChans]
            if any(eegChansInArea_Bool):
                eegChansInArea = [x for (x,y) in zip(eegChans, eegChansInArea_Bool) if y]
                print('Loading LFP data for channels: ' +  ', '.join(map(str, eegChansInArea)))
                eegData = []
                for OpenEphysDataPath in OpenEphysDataPaths:
                    print([hfunct.time_string(), 'DEBUG: loading eegData for ', OpenEphysDataPath])
                    eegData.append(load_eegData(OpenEphysDataPath, eegChansInArea))
                print([hfunct.time_string(), 'DEBUG: concatenating eeg data'])
                eegData = concatenate_eegData_across_recordings(eegData, data_time_edges, 
                                                                recording_edges)
            else:
                eegData = None
        createAxonaData(AxonaDataPath, spike_data, combined_data_time_edges, 
                        posdata=posdata, experiment_info=experiment_info, 
                        axona_file_name=area, eegData=eegData, 
                        pixels_per_metre=pixels_per_metre, show_output=show_output)
    with open(os.path.join(AxonaDataPath, 'recording_edges'), 'w') as file:
        for edges, OpenEphysDataPath in zip(recording_edges, OpenEphysDataPaths):
            file.write(str(edges) + ' path: ' + OpenEphysDataPath + '\n')

def createAxonaData(AxonaDataPath, spike_data, data_time_edges, posdata=None, 
                    experiment_info=None, axona_file_name='datafile', eegData=None, 
                    pixels_per_metre=None, show_output=False):
    '''
    AxonaDataPath - folder path where to store AxonaData files
    spike_data - see create_DACQ_waveform_data() for description
    posdata - create_DACQ_pos_data() for description
    data_time_edges - beginning and end time of data to convert to AxonaData. Rest is discarded.
                      posdata and eegData, if provided, must cover this range entirely.
    experiment_info - see getExperimentInfo() for description
    axona_file_name - str - filename prefix for output AxonaData files
    eegData - see create_DACQ_eeg_or_egf_data() for description
    pixels_per_metre - see create_DACQ_pos_data() for description
    show_output - bool - if True, destination folder is opened with xdg-open when finished

    Note! - timestamps info in spike_data, posdata and eegData must be aligned.
    '''
    n_tetrodes = len(spike_data)
    # Convert data to DACQ format
    print('Converting spike data')
    waveform_data_dacq = create_DACQ_waveform_data(spike_data, data_time_edges)
    print('Converting position data')
    pos_data_dacq, pixels_per_metre, limits = create_DACQ_pos_data(posdata, data_time_edges, 
                                                                   pixels_per_metre)
    if not (eegData is None):
        print('Converting LFP to EEG data')
        eeg_data_dacq = create_DACQ_eeg_data(eegData, data_time_edges)
        print('Converting LFP to EGF data')
        egf_data_dacq = create_DACQ_egf_data(eegData, data_time_edges)
    else:
        EEG_samples_per_position = (data_time_edges[1] - data_time_edges[0]) / len(pos_data_dacq)
        num_EEG_samples = int(np.round(AxonaDataEEG_SamplingRate() * (data_time_edges[1] - data_time_edges[0])))
        num_EGF_samples = int(np.round(AxonaDataEGF_SamplingRate() * (data_time_edges[1] - data_time_edges[0])))
    # Get recording specific header data for both datatypes
    if experiment_info is None:
        experiment_info = getExperimentInfo()
    experiment_info['duration'] = str(data_time_edges[1] - data_time_edges[0])
    nvds = {'num_spikes': []}
    num_spikes = []
    for ntet in range(n_tetrodes):
        nvds['num_spikes'].append(str(int(len(waveform_data_dacq[ntet]) / 4)))
    nvds['num_pos_samples'] = str(len(pos_data_dacq))
    nvds['pixels_per_metre'] = str(pixels_per_metre)
    if not (eegData is None):
        nvds['EEG_samples_per_position'] = str(int(np.round(len(eeg_data_dacq[0]) / len(pos_data_dacq))))
        nvds['num_EEG_samples'] = str(len(eeg_data_dacq[0]))
        nvds['num_EGF_samples'] = str(len(egf_data_dacq[0]))
    else:
        nvds['EEG_samples_per_position'] = EEG_samples_per_position
        nvds['num_EEG_samples'] = num_EEG_samples
        nvds['num_EGF_samples'] = num_EGF_samples
    # Get waveform file headers templates and update with recording specific data
    headers_wave = []
    for ntet in range(n_tetrodes):
        header_wave, keyorder_wave = header_templates('waveforms')
        new_values_dict = {'num_spikes': nvds['num_spikes'][ntet]}
        header_wave = update_header(header_wave, experiment_info, new_values_dict)
        headers_wave.append(header_wave)
    # Get position file header templates and update with recording specific data
    header_pos, keyorder_pos = header_templates('pos')
    new_values_dict = {'num_pos_samples': nvds['num_pos_samples'], 
                       'pixels_per_metre': nvds['pixels_per_metre'], 
                       'max_x': str(int(limits['x_max'])), 
                       'max_y': str(int(limits['y_max'])), 
                       'window_max_x': str(int(limits['x_max'])), 
                       'window_max_y': str(int(limits['y_max'])), 
                       'min_x': str(int(limits['x_min'])), 
                       'min_y': str(int(limits['y_min'])), 
                       'window_min_x': str(int(limits['x_min'])), 
                       'window_min_y': str(int(limits['y_min']))}
    header_pos = update_header(header_pos, experiment_info, new_values_dict)
    # Get EEG file header templates and update with recording specific data
    header_eeg, keyorder_eeg = header_templates('eeg')
    new_values_dict = {'EEG_samples_per_position': nvds['EEG_samples_per_position'], 
                       'num_EEG_samples': nvds['num_EEG_samples']}
    header_eeg = update_header(header_eeg, experiment_info, new_values_dict)
    # Get EGF file header templates and update with recording specific data
    header_egf, keyorder_egf = header_templates('egf')
    new_values_dict = {'num_EGF_samples': nvds['num_EGF_samples']}
    header_egf = update_header(header_egf, experiment_info, new_values_dict)
    # Create AxonaDataPath directory if it does not exist
    if not os.path.exists(AxonaDataPath):
        os.mkdir(AxonaDataPath)
    # Write WAVEFORM data for each tetrode into DACQ format
    print('Writing files to disk')
    for ntet in range(n_tetrodes):
        fname = os.path.join(AxonaDataPath, axona_file_name + '.' + str(ntet + 1))
        write_file_in_axona_format(fname, headers_wave[ntet], keyorder_wave, waveform_data_dacq[ntet])
    # Write POSITION data into DACQ format
    fname = os.path.join(AxonaDataPath, axona_file_name + '.pos')
    write_file_in_axona_format(fname, header_pos, keyorder_pos, pos_data_dacq)
    # Write EEG data into DACQ format
    if not (eegData is None):
      for i, eeg_data in enumerate(eeg_data_dacq):
          fname = os.path.join(AxonaDataPath, axona_file_name + '.eeg')
          if i > 0:
              fname += str(i + 1)
          write_file_in_axona_format(fname, header_eeg, keyorder_eeg, eeg_data)
    # Write EGF data into DACQ format
    if not (eegData is None):
      for i, egf_data in enumerate(egf_data_dacq):
          fname = os.path.join(AxonaDataPath, axona_file_name + '.egf')
          if i > 0:
              fname += str(i + 1)
          write_file_in_axona_format(fname, header_egf, keyorder_egf, egf_data)
    # Write CLU files
    for ntet in range(n_tetrodes):
        cluFileName = os.path.join(AxonaDataPath, axona_file_name + '.clu.' + str(ntet + 1))
        if (isinstance(spike_data[ntet]['clusterIDs'], np.ndarray) 
                and len(spike_data[ntet]['clusterIDs']) > 0):
            write_clusterIDs_in_CLU_format(spike_data[ntet]['clusterIDs'], cluFileName)
    # Write SET file
    setFileName = os.path.join(AxonaDataPath, axona_file_name + '.set')
    duration_string = experiment_info['duration'] + (9 - len(experiment_info['duration'])) * ' '
    new_values_dict = {'duration': duration_string}
    write_set_file(setFileName, new_values_dict)
    # Opens recording folder with Ubuntu file browser
    if show_output:
        subprocess.Popen(['xdg-open', AxonaDataPath])
    print('Finished creating AxonaData.')

# The following is the default ending for a QtGui application script
if __name__ == "__main__":
    # Input argument handling and help info
    parser = argparse.ArgumentParser(description='Export data into Axona format.')
    parser.add_argument('paths', type=str, nargs='*', 
                        help='recording data folder(s) (can enter multiple paths separated by spaces to KlustaKwik simultaneously)')
    parser.add_argument('--chan', type=int, nargs = 2, 
                        help='list the first and last channel to process (counting starts from 1)')
    parser.add_argument('--eegChans', type=int, nargs = '*', 
                        help='enter channel number to use for creating EEG data')
    parser.add_argument('--subfolder', type=str, nargs = 1, 
                        help='enter the name of subfolder to use for AxonaData')
    parser.add_argument('--spike_name', type=str, nargs = 1, 
                        help='enter the name of spike set in NWB file to use for AxonaData')
    parser.add_argument('--ppm', type=int, nargs = 1, 
                        help='enter pixels_per_metre to assume position data is in pixels')
    parser.add_argument('--concatenatedDataPath', type=str, nargs = 1, 
                        help='enter the path to folder where to write data concatenated from all specified recordings')
    parser.add_argument('--show_output', action='store_true', 
                        help='to open AxonaData output folder after processing')
    parser.add_argument('--clustering_name', type=str, nargs = 1, 
                        help='specify cluster identities path to use in NWB file')
    args = parser.parse_args()
    # Get paths to recording files
    OpenEphysDataPaths = args.paths
    if isinstance(OpenEphysDataPaths, str):
        OpenEphysDataPaths = [OpenEphysDataPaths]
    # If directories entered as paths, attempt creating path to file by appending experiment_1.nwb
    for ndata, OpenEphysDataPath in enumerate(OpenEphysDataPaths):
        if not os.path.isfile(OpenEphysDataPath):
            new_path = os.path.join(OpenEphysDataPath, 'experiment_1.nwb')
            if os.path.isfile(new_path):
                OpenEphysDataPaths[ndata] = new_path
            else:
                raise ValueError('The following path does not lead to a NWB data file:\n' + OpenEphysDataPath)
    # Get chan input variable
    if args.chan:
        chan = [args.chan[0] - 1, args.chan[1]]
        if np.mod(chan[1] - chan[0], 4) != 0:
            raise ValueError('Channel range must cover full tetrodes')
        area_name = 'Chan' + str(args.chan[0]) + '-' + str(args.chan[1])
        channel_map = {area_name: {'list': list(range(chan[0], chan[1], 1))}}
    else:
        channel_map = None
    # Get eegChans variable
    if args.eegChans:
        eegChans = [x - 1 for x in args.eegChans]
    else:
        eegChans = None
    # Get subfolder variable
    if args.subfolder:
        subfolder = args.subfolder[0]
    else:
        subfolder = 'AxonaData'
    # Get spike_name variable
    if args.spike_name:
        spike_name = args.spike_name[0]
    else:
        spike_name = 'first_available'
    # Get ppm variable
    if args.ppm:
        pixels_per_metre = args.ppm[0]
    else:
        pixels_per_metre = None
    # Get concatenatedDataPath variable
    if args.concatenatedDataPath:
        concatenatedDataPath = args.concatenatedDataPath[0]
    else:
        concatenatedDataPath = None
    # Get show_output variable
    if args.show_output:
        show_output = True
    else:
        show_output = False
    # Get clustering_name variable
    if args.clustering_name:
        clustering_name = args.clustering_name[0]
    else:
        clustering_name = None
    # Run main script for each NWB file or to concatenate them if requested
    if concatenatedDataPath is None:
        for OpenEphysDataPath in OpenEphysDataPaths:
            createAxonaData_for_NWBfile(OpenEphysDataPath, spike_name=spike_name, 
                                        channel_map=channel_map, subfolder=subfolder, 
                                        eegChans=eegChans, pixels_per_metre=pixels_per_metre, 
                                        show_output=show_output, clustering_name=clustering_name)
    else:
        createAxonaData_for_multiple_NWBfiles(OpenEphysDataPaths, concatenatedDataPath, 
                                              spike_name=spike_name, channel_map=channel_map, 
                                              eegChans=eegChans, pixels_per_metre=pixels_per_metre, 
                                              show_output=show_output, clustering_name=clustering_name)
